/*
 * Copyright (c) 2000-2003 QoSient, LLC
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * argus client library
 *
 * written by Carter Bullard
 * QoSient, LLC
 *
 */


#ifndef ArgusClient
#define ArgusClient
#endif

#include <argus_client.h>
#include <math.h>

#define RA_ADD_OPTION		1
#define RA_SUB_OPTION		2

void RaProcessSOptions(void);

void
RaProcessSOptions()
{
   int i, x, RaOptionOperation, RaOptionRank;
   char *soption = NULL;

   if ((soption = RaSOptionStrings[0]) != NULL) {
      if (!((*soption == '+') || (*soption == '-')))
         bzero ((char *) RaPrintAlgorithms, sizeof(RaPrintAlgorithms));
   } else
      ArgusLog (LOG_ERR, "ArgusClientInit Index %d String NULL", RaSOptionIndex);

   for (i = 0; i < ARGUS_MAX_S_OPTIONS; i++) {
      if ((soption = RaSOptionStrings[i]) != NULL) {
         RaOptionOperation = RA_ADD_OPTION;
         RaOptionRank = -1;
         if ((*soption == '+') || (*soption == '-')) {
            if (*soption == '-')
               RaOptionOperation = RA_SUB_OPTION;

            soption++;
            if (isdigit((int)*soption)) {
               sscanf(soption, "%d", &RaOptionRank);
               while(isdigit((int)*soption)) soption++;
            }
         }

         RaPrintDate = 0;
         for (x = 0; x < MAX_PRINT_ALG_TYPES; x++) {
            if (!strncmp (RaPrintKeyWords[x], soption, strlen(RaPrintKeyWords[x]))) {
	       if (x == 0) {
		  RaPrintStartTime = 1;
		  RaPrintLastTime = 1;
	       }
               if (x < 3) RaPrintDate++;
               switch (RaOptionOperation) {
                  case RA_ADD_OPTION:
                     if (RaOptionRank == -1) {
                        int z = 0;
                        while (RaPrintAlgorithms[z] != NULL) z++;
                        if (z < ARGUS_MAX_PRINT_ALG)
                           RaPrintAlgorithms[z] = RaPrintAlgorithmTable[x];
                     } else {
                        int z = RaOptionRank;
                        while (RaPrintAlgorithms[z] != NULL) z++;
                        while (z != RaOptionRank) {
                           RaPrintAlgorithms[z] = RaPrintAlgorithms[z - 1];
                           RaPrintAlgorithms[z - 1] = NULL;
                           z--;
                        }
                        RaPrintAlgorithms[z] = RaPrintAlgorithmTable[x];
                     }
                     break;

                  case RA_SUB_OPTION: {
                     for (RaOptionRank = 0; RaOptionRank < ARGUS_MAX_PRINT_ALG; RaOptionRank++)
                        if (RaPrintAlgorithms[RaOptionRank] == RaPrintAlgorithmTable[x])
                           break;

                     if (RaOptionRank < ARGUS_MAX_PRINT_ALG) {
                        RaPrintAlgorithms[RaOptionRank] = NULL;

                        while (RaOptionRank < ARGUS_MAX_PRINT_ALG) {
                           RaPrintAlgorithms[RaOptionRank] = RaPrintAlgorithms[RaOptionRank + 1];
                           RaOptionRank++;
                        }
                     }
                     break;
                  }
               }
               break;
            }
         }

      } else
         break;
   }
}

struct ArgusRecord *
RaCopyArgusRecord (struct ArgusRecord *argus)
{
   int length = 0;
   struct ArgusRecord *retn = NULL;

   if (argus && ((length = argus->ahdr.length) > 0))
      if ((retn = (struct ArgusRecord *) ArgusCalloc(1, length)) != NULL)
         bcopy ((char *) argus, (char *) retn, length);

   return (retn);
}


long long
RaGetActiveDuration (struct ArgusRecord *argus)
{
   return (RaGetuSecDuration (argus));
}

long long RaGetuSecAvgDuration (struct ArgusRecord *);

long long
RaGetuSecAvgDuration (struct ArgusRecord *argus)
{
   struct ArgusAGRStruct *agr;
   struct ArgusFarHeaderStruct *argusthisfarhdrs[32];
   long long retn = 0;
   int sec, usec;

   ArgusIndexRecord (argus,  argusthisfarhdrs);

   if ((agr = (struct ArgusAGRStruct *) argusthisfarhdrs[ARGUS_AGR_DSR_INDEX]) != NULL) {
      int ArgusThisMultiplier = 1000;

      if (agr->status & ARGUS_AGR_USECACTTIME)
         ArgusThisMultiplier = 1;

      if (agr->act.n)
         return (agr->act.meanval * ArgusThisMultiplier);
   }

   sec  = argus->argus_far.time.last.tv_sec  - argus->argus_far.time.start.tv_sec;
   usec = argus->argus_far.time.last.tv_usec - argus->argus_far.time.start.tv_usec;

   if (usec < 0) {
      sec--; usec += 1000000;
   }

   retn = ((long long) (sec * 1000000)) + (long long) usec;

   return (retn);
}

long long
RaGetuSecDuration (struct ArgusRecord *argus)
{
   unsigned int retn = 0;
   int sec, usec;

   sec  = argus->argus_far.time.last.tv_sec  - argus->argus_far.time.start.tv_sec;
   usec = argus->argus_far.time.last.tv_usec - argus->argus_far.time.start.tv_usec;

   if (usec < 0) {
      sec--; usec += 1000000;
   }

   retn = ((long long)(sec * 1000000)) + (long long)usec;

   return (retn);
}

int
RaTypeMatches (u_short policytype, u_short recordtype)
{
   int retn = 1;

   if (policytype != 0xFFFF) 
      if ((policytype != recordtype) && (recordtype != 65535))
         retn = 0;

   return(retn);
}


int
RaCIDRAddrMatches (u_int addr, struct RaCIDRAddr *cidr)
{
   int retn = 1;

   switch (cidr->type) {

      case RA_CIDR_TYPE:
         if (cidr->addr) {
            if ((addr & cidr->mask) != cidr->addr)
               retn = 0;
         }
         break;

      case RA_NETCLASS_TYPE:
         if (cidr->addr) {
            if ((addr & ipaddrtonetmask(addr)) != cidr->addr)
               retn = 0;
         }
         break;
   }

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaCIDRAddrMatches(0x%x, 0x%x) returned %d\n", addr, cidr, retn);
#endif

   return (retn);
}

int
RaProtoMatches (u_short p, int proto)
{
   int retn = 1;

   if (proto != 0xFF)
      retn = (p == proto) ? 1 : 0;

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaProtoMatches (%d, %d) returned %d\n", p, proto, retn);
#endif

   return (retn);
}


int
RaPortMatches (u_short p1, u_short p2)
{
   int retn = 1;
   
   if (p2 != 0xFFFF)
      retn = (p1 == p2) ? 1 : 0;

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaPortMatches (%d, %d) returned %d\n", p1, p2, retn);
#endif

   return (retn);
}

int
RaPolicyMatch (struct ArgusRecord *argus, struct RaPolicyStruct *rap)
{
   int retn = 0;
   struct ArgusFlow *flow = &argus->argus_far.flow;
   u_char proto = flow->ip_flow.ip_p;

   if (RaTypeMatches ((argus->ahdr.status & 0xFFFF), rap->type))
   if (RaCIDRAddrMatches (flow->ip_flow.ip_src, &rap->src))
   if (RaCIDRAddrMatches (flow->ip_flow.ip_dst, &rap->dst))
   if (RaProtoMatches (flow->ip_flow.ip_p, rap->proto))

   switch (proto) {
      case IPPROTO_TCP:
      case IPPROTO_UDP:
         if (RaPortMatches (flow->ip_flow.sport, rap->sport))
         if (RaPortMatches (flow->ip_flow.dport, rap->dport))
            retn = 1;
         break;

      case IPPROTO_ICMP:
         if (RaPortMatches (flow->icmp_flow.type, rap->sport))
         if (RaPortMatches (flow->icmp_flow.code, rap->dport))
         retn = 1;
         break;

      default:
         retn = 1;
         break;
   }


#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaPolicyMatch(0x%x, 0x%x) returned %d\n", argus, rap, retn);
#endif

   return (retn);
}


struct RaPolicyStruct *
RaFlowModelOverRides (struct ArgusRecord *argus, struct RaFlowModelStruct *fmodel)
{
   struct RaPolicyStruct *retn = NULL;
   int i;

   if (fmodel) {
      switch (argus->ahdr.status & 0xFFFF) {
         case ETHERTYPE_ARP:
         case ETHERTYPE_IP:
            for (i = 0; i < fmodel->pindex; i++) {
               if (RaPolicyMatch (argus, fmodel->policy[i])) {
                  retn = fmodel->policy[i];
                  break;
               }
            }
            break;

         case ETHERTYPE_REVARP:
            break;

         default:
            break;
      }
   }

#ifdef ARGUSDEBUG
   ArgusDebug (7, "RaFlowModelOverRides(0x%x) returned 0x%x\n", argus, retn);
#endif

   return (retn);
}

#if defined(__OpenBSD__)
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#endif

#include <netinet/ip_icmp.h>
#include <netinet/igmp.h>

void
RaModifyDefaultFlow (struct ArgusRecord *argus)
{
   struct ArgusFlow *flow = &argus->argus_far.flow;

   switch (argus->ahdr.status & 0xFFFF) {
      case ETHERTYPE_IP:
         if (argus->argus_far.flow.ip_flow.tp_p == ARGUS_FRAG_FLOWTAG)
            argus->argus_far.flow.ip_flow.tp_p = 0;

         switch (argus->argus_far.flow.ip_flow.ip_p) {
            case IPPROTO_UDP:
            case IPPROTO_TCP:
               break;

            case IPPROTO_ICMP:
               if (flow->icmp_flow.type != ICMP_UNREACH)
                  flow->icmp_flow.id = 0xFFFF;

               flow->icmp_flow.ip_id = 0xFFFF;
               break;

            default:
               break;
         }
         break;

      case ETHERTYPE_ARP:
      case ETHERTYPE_REVARP:
         break;

      default:
         break;
   }
}

int RaFlowMajorModified = 0;

void
RaModifyFlow (struct RaPolicyStruct *rap, struct ArgusRecord *argus, struct RaFlowModelStruct *fmodel)
{
   struct RaPolicyStruct *model = NULL;
   struct ArgusFlow *flow = &argus->argus_far.flow;
   u_short proto;

   RaFlowMajorModified = 0;

   if (fmodel->model) {
      int i;
      for (i = 0; i < fmodel->mindex; i++) {
         if (fmodel->model[i]->RaPolicyId == rap->RaModelId) {
            model = fmodel->model[i];
            break;
         }
      }
   } else
      ArgusLog (LOG_ERR, "RaModifyFlow () model is NULL");

   switch (argus->ahdr.status & 0xFFFF) {
      case ETHERTYPE_IP:
         if (flow->ip_flow.tp_p == ARGUS_FRAG_FLOWTAG) {
            flow->ip_flow.tp_p = 0;
         }

         if (model) {
            switch (model->src.type) {
               case RA_CIDR_TYPE:
                  flow->ip_flow.ip_src = flow->ip_flow.ip_src & model->src.addr;
                  break;

               case RA_NETCLASS_TYPE:
                  flow->ip_flow.ip_src = flow->ip_flow.ip_src & ipaddrtonetmask(flow->ip_flow.ip_src);
                  break;
            }

            switch (model->dst.type) {
               case RA_CIDR_TYPE:
                  flow->ip_flow.ip_dst = flow->ip_flow.ip_dst & model->dst.addr;
                  break;

               case RA_NETCLASS_TYPE:
                  flow->ip_flow.ip_dst = flow->ip_flow.ip_dst & ipaddrtonetmask(flow->ip_flow.ip_dst);
                  break;
            }

            proto = flow->ip_flow.ip_p;
            flow->ip_flow.ip_p  &= model->proto;

            switch (proto) {
               case IPPROTO_TCP:
                  flow->ip_flow.tp_p = 0;

               case IPPROTO_UDP:
                  switch (proto) {
                     case IPPROTO_TCP:
                        if (flow->ip_flow.sport == 20) {
                           if (!(model->sport))
                              flow->ip_flow.dport = 0xFFFF;
                           if (!(model->dport)) {
                              flow->ip_flow.sport = 0xFFFF;
                              RaFlowMajorModified = 1;
                           }
                        } else

                     case IPPROTO_UDP:
                        if (!(model->sport))
                           flow->ip_flow.sport = 0xFFFF;
                        if (!(model->dport)) {
                           flow->ip_flow.dport = 0xFFFF;
                           RaFlowMajorModified = 1;
                        }
                        break;

                     default:
                        flow->ip_flow.sport = 0xFFFF;
                        flow->ip_flow.dport = 0xFFFF;
                        break;
                  }
                  break;
          
               case IPPROTO_ICMP:
                  if (!(model->sport)) {
                     flow->icmp_flow.type = 0xFF;
                     RaFlowMajorModified = 1;
                  }

                  if (!(model->dport)) {
                     flow->icmp_flow.code = 0xFF;
                     RaFlowMajorModified = 1;
                  }

                  flow->icmp_flow.id = 0xFFFF;
                  flow->icmp_flow.ip_id = 0xFFFF;
                  flow->icmp_flow.tp_p = 0;
                  break;

               default:
                  flow->ip_flow.tp_p = 0;
                  if (!(model->sport))
                     flow->ip_flow.sport = 0xFFFF;
                  else
                     flow->ip_flow.sport = 0;
                  if (!(model->dport))
                     flow->ip_flow.dport = 0xFFFF;
                  else
                     flow->ip_flow.dport = 0;
                  RaFlowMajorModified = 1;
                  break;
            }

            if (flow->ip_flow.ip_p == 0)
               flow->ip_flow.tp_p  = 0;
         }
         break;

      case ETHERTYPE_ARP:
         if (model) {
            switch (model->src.type) {
               case RA_CIDR_TYPE:
                  flow->arp_flow.arp_spa = flow->arp_flow.arp_spa & model->src.addr;
                  break;

               case RA_NETCLASS_TYPE:
                  flow->arp_flow.arp_spa = flow->arp_flow.arp_spa & ipaddrtonetmask(flow->ip_flow.ip_src);
                  break;
            }

            switch (model->dst.type) {
               case RA_CIDR_TYPE:
                  flow->arp_flow.arp_tpa = flow->arp_flow.arp_tpa & model->dst.addr;
                  break;

               case RA_NETCLASS_TYPE:
                  flow->arp_flow.arp_tpa = flow->arp_flow.arp_tpa & ipaddrtonetmask(flow->ip_flow.ip_dst);
                  break;
            }

            bzero (flow->arp_flow.etheraddr, 6);
            flow->arp_flow.pad = 0;
         }

         break;

      case ETHERTYPE_REVARP:
      default:
         break;
   }

#ifdef ARGUSDEBUG
   ArgusDebug (7, "RaModifyFlow(0x%x, 0x%x) done.\n", rap, argus);
#endif
}


int RaThisFlowRev = 0;

struct ArgusRecordStore *
RaFindRevArgusRecord(struct RaHashTableStruct *htable, struct ArgusRecord *argus)
{
   struct ArgusRecordStore *retn = NULL;
   struct RaHashTableHeader *hashEntry;
   struct ArgusFlow flowbuf, *RaThisFlow = &flowbuf;
   struct RaHashStruct ArgusHash;
   struct ArgusFlow flow;
   unsigned short *ptr = NULL;
   unsigned short ip_id = 0;
   int i, len;

   bzero ((char *)&ArgusHash, sizeof(ArgusHash));
   ArgusHash.len = sizeof(ArgusHash.buf);

   *RaThisFlow = argus->argus_far.flow;
   flow = *RaThisFlow;
   RaThisFlowRev = 0;

   if (!(RaAutoCorrection))
      return (retn);

   if ((argus->ahdr.status & 0xFFFF) == ETHERTYPE_IP) {
      switch (flow.ip_flow.ip_p) {
         case IPPROTO_UDP:              
         case IPPROTO_TCP:              
            RaThisFlow->ip_flow.ip_src = flow.ip_flow.ip_dst;
            RaThisFlow->ip_flow.ip_dst = flow.ip_flow.ip_src;
            RaThisFlow->ip_flow.sport  = flow.ip_flow.dport;
            RaThisFlow->ip_flow.dport  = flow.ip_flow.sport;
            ip_id = RaThisFlow->ip_flow.ip_id;
            RaThisFlow->ip_flow.ip_id = 0;
            break;

         case IPPROTO_ICMP: {
            RaThisFlow->icmp_flow.ip_src = flow.icmp_flow.ip_dst;
            RaThisFlow->icmp_flow.ip_dst = flow.icmp_flow.ip_src;
            ip_id = RaThisFlow->icmp_flow.ip_id;
            RaThisFlow->icmp_flow.ip_id = 0;

            switch (RaThisFlow->icmp_flow.type) {
               case ICMP_MASKREQ:
                  RaThisFlow->icmp_flow.type = ICMP_MASKREPLY;
                  break;
               case ICMP_MASKREPLY:
                  RaThisFlow->icmp_flow.type = ICMP_MASKREQ;
                  break;
               case ICMP_ECHO:
                  RaThisFlow->icmp_flow.type = ICMP_ECHOREPLY;
                  break;
               case ICMP_ECHOREPLY:
                  RaThisFlow->icmp_flow.type = ICMP_ECHO;
                  break;
               case ICMP_TSTAMP:
                  RaThisFlow->icmp_flow.type = ICMP_TSTAMPREPLY;
                  break;
               case ICMP_TSTAMPREPLY:
                  RaThisFlow->icmp_flow.type = ICMP_TSTAMP;
                  break;
               case ICMP_IREQ:
                  RaThisFlow->icmp_flow.type = ICMP_IREQREPLY;
                  break;
               case ICMP_IREQREPLY:
                  RaThisFlow->icmp_flow.type = ICMP_IREQ;
                  break;
            }
            break;
         }

         case IPPROTO_ESP:
         default:
            RaThisFlow->ip_flow.ip_src = flow.ip_flow.ip_dst;
            RaThisFlow->ip_flow.ip_dst = flow.ip_flow.ip_src;
            ip_id = RaThisFlow->ip_flow.ip_id;
            RaThisFlow->ip_flow.ip_id = 0;
            break;
      }
   }

   ArgusHash.buf[0] = argus->ahdr.argusid;
   bcopy (RaThisFlow, (char *)&ArgusHash.buf[1], sizeof(*RaThisFlow));
   ptr = (unsigned short *)&ArgusHash.buf[0];

   for (i = 0, len = ArgusHash.len / sizeof(unsigned short); i < len; i++)
      ArgusHash.hash += *ptr++;

   if ((hashEntry = RaFindHashEntry (htable, &ArgusHash)) != NULL) {
      RaThisFlowRev++;
      retn = hashEntry->storeobj;
   }

   switch (argus->ahdr.status & 0xFFFF) {
      case ETHERTYPE_IP: {
         switch (flow.ip_flow.ip_p) {
            default:
            case IPPROTO_TCP:
            case IPPROTO_UDP:
               RaThisFlow->ip_flow.ip_id = ip_id;
               break;

            case IPPROTO_ICMP:
               RaThisFlow->icmp_flow.ip_id = ip_id;
               break;
         }
      }
   }

   return (retn);
}


struct ArgusRecordStore *
RaFindArgusRecord(struct RaHashTableStruct *htable, struct ArgusRecord *argus)
{
   struct ArgusRecordStore *retn = NULL;
   struct RaHashTableHeader *hashEntry;
   struct ArgusFlow *RaThisFlow = NULL;
   unsigned short *ptr, ip_id = 0;

   struct RaHashStruct ArgusHash;
   int i, len;

   bzero ((char *)&ArgusHash, sizeof(ArgusHash));
   ArgusHash.len = sizeof(ArgusHash.buf);

   RaThisFlowRev = 0;
   RaThisFlow = &argus->argus_far.flow;

   switch (argus->ahdr.status & 0xFFFF) {
      case ETHERTYPE_IP: {
         switch (argus->argus_far.flow.ip_flow.ip_p) {
            case IPPROTO_ICMP:
               ip_id = RaThisFlow->icmp_flow.ip_id;
               RaThisFlow->icmp_flow.ip_id = 0;
               break;

            default:
               ip_id = RaThisFlow->ip_flow.ip_id;
               RaThisFlow->ip_flow.ip_id = 0;
               break;
         }
      }
   }

   ArgusHash.buf[0] = argus->ahdr.argusid;
   bcopy (RaThisFlow, (char *)&ArgusHash.buf[1], sizeof(*RaThisFlow));
   ptr = (unsigned short *)&ArgusHash.buf[0];

   for (i = 0, len = ArgusHash.len / sizeof(unsigned short); i < len; i++)
      ArgusHash.hash += *ptr++;

   if ((hashEntry = RaFindHashEntry (htable, &ArgusHash)) != NULL)
      retn = hashEntry->storeobj;

   switch (argus->ahdr.status & 0xFFFF) {
      case ETHERTYPE_IP: {
         switch (argus->argus_far.flow.ip_flow.ip_p) {

            case IPPROTO_ICMP:
               RaThisFlow->icmp_flow.ip_id = ip_id;
               break;

            default:
               RaThisFlow->ip_flow.ip_id = ip_id;
               break;
         }
      }
   }

#ifdef ARGUSDEBUG
   ArgusDebug (4, "RaFindArgusRecord: returning 0x%x\n", retn);
#endif

   return (retn);
}


struct ArgusRecordStore *
RaNewArgusStore(struct ArgusRecord *argus)
{
   struct ArgusRecordStore *retn = NULL;
   struct timeval lasttime;

   if ((retn = (struct ArgusRecordStore *) ArgusCalloc (1, sizeof(struct ArgusRecordStore))) != NULL) {
      RaAllocArgusRecordStore++;
      if (argus != NULL) {
         lasttime = argus->argus_far.time.last;

         retn->qhdr.lasttime = lasttime;
         retn->qhdr.logtime  = lasttime;
      }
      if ((retn->data = (struct ArgusRecordData **) ArgusCalloc (RaHistoTimeSeries + 1, sizeof(void *))) == NULL)
         ArgusLog (LOG_ERR, "RaNewArgusStore () ArgusCalloc %s\n", strerror(errno));
   } else
     ArgusLog (LOG_ERR, "RaNewArgusStore () ArgusCalloc %s\n", strerror(errno));

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaNewArgusStore: added 0x%x\n", retn);
#endif

   return (retn);
}

void RaDeleteArgusStore(struct ArgusRecordStore *);
void RaDeleteArgusData(struct ArgusRecordData *);

/*
struct ArgusRecordStore {
   struct ArgusQueueHeader qhdr;
   struct RaHashTableStruct *htable;
   struct RaHashTableHeader *rahtblhdr;
   struct RaQueueStruct *queue;
   struct timeval startime, lasttime;
   int  ArgusTimeout, ArgusIdleTimeout;
   int  status, sortag;
   char *displaystr;
   struct ArgusRecordData **data;
   struct ArgusRecord *cor[8];
   struct RaSrvSignature *sig;
   struct ArgusStoreObjectStruct *object;
};
*/

void
RaDeleteArgusStore(struct ArgusRecordStore *store)
{
   int i = 0;

   if (store->qhdr.queue != NULL)
      RaRemoveFromQueue(store->qhdr.queue, &store->qhdr);

   if (store->rahtblhdr != NULL) {
      if (store->rahtblhdr->htable != NULL)
         RaRemoveHashEntry (store->rahtblhdr->htable, store->rahtblhdr);

      store->rahtblhdr = NULL;
   }

   if (store->data != NULL) {
      for (i = 0; i < RaHistoTimeSeries + 1; i++)
         if (store->data[i] != NULL)
            RaDeleteArgusData(store->data[i]);

      ArgusFree(store->data);
      store->data = NULL;
   }

   if (store->queue != NULL) {
      RaDeleteQueue(store->queue);
      store->queue = NULL;
   }

   if (store->displaystr != NULL) {
      free(store->displaystr);
      store->displaystr = NULL;
   }

   if (store->object != NULL) {
      ArgusFree(store->object);
      store->object = NULL;
   }
   
   ArgusFree(store);

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaDeleteArgusStore(0x%x): Done\n", store);
#endif
}

int RaAllocArgusRecordData = 0;
int RaTotalDataRecords = 0;

struct ArgusRecordData *
RaNewArgusData(struct ArgusRecord *argus)
{
   struct ArgusRecordData *retn = NULL;
   struct timeval lasttime;

   if ((retn = (struct ArgusRecordData *) ArgusCalloc (1, sizeof (*retn))) != NULL) {
      RaAllocArgusRecordData++;
   
      if ((retn->argus = RaCopyArgusRecord(argus)) != NULL) {
         RaTotalDataRecords++;

         retn->agr.laststartime = argus->argus_far.time.start;
         lasttime = argus->argus_far.time.last;

         retn->agr.type     = ARGUS_AGR_DSR;
         retn->agr.length   = sizeof(retn->agr);
         retn->agr.lasttime = lasttime;
         retn->agr.act.maxval  = RaThisActiveDuration;
         retn->agr.act.minval  = RaThisActiveDuration;

         retn->act.sumtime = RaThisActiveDuration;

         if (RaThisActiveDuration != 0)
            retn->act.sumsqrd  = pow (RaThisActiveDuration, 2.0);
         retn->act.n        = 1;
         retn->agr.count    = 1;

         retn->agr.idle.minval = 0x7FFFFFFF;
         retn->agr.status  |= ARGUS_AGR_USECACTTIME;
         retn->agr.status  |= ARGUS_AGR_USECIDLETIME;

         bcopy ((char *) RaArgusFlow, (char *)&retn->flow, sizeof(retn->flow));

      } else {
         ArgusFree (retn);
         RaAllocArgusRecordData--;
         retn = NULL;
      }
   }

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaNewArgusData: added 0x%x\n", retn);
#endif

   return (retn);
}

void
RaDeleteArgusData(struct ArgusRecordData *data)
{
   if (data->argus != NULL)
      ArgusFree(data->argus);

   ArgusFree(data);

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaRemoveArgusRecord: done\n");
#endif
}

void
RaRemoveArgusRecord(struct ArgusRecord *argus)
{

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaRemoveArgusRecord: done\n");
#endif
}


void
RaMergeArgusRecord(struct ArgusRecord *a1, struct ArgusRecordStore *store, int index)
{
   u_char buf[MAXSTRLEN];
   struct ArgusRecordData *data = store->data[index];
   struct ArgusRecord *a2 = NULL, *tmp = (struct ArgusRecord *)buf;
   struct ArgusFarHeaderStruct *a1farhdr[32], *a2farhdr[32];
   struct ArgusFarStruct *far1 = NULL;
   struct ArgusFarStruct *far2 = NULL;
   struct ArgusFarStruct *ArgusThisDataFar;
   unsigned int a1DSRStatus = 0, a2DSRStatus = 0;
   unsigned int i, status = 0, TimeMultiplier;
   struct ArgusAGRStruct *agr;
   unsigned int duration, rev = 0;

   if (!(a1 && data))
      return;

   if ((a2 = data->argus) != NULL) {
      bzero (buf, sizeof(buf));
      bcopy ((char *) a2, buf, a2->ahdr.length);

      ArgusThisDataFar = (struct ArgusFarStruct *) data->farhdrs[ARGUS_FAR_DSR_INDEX];

      a1DSRStatus = ArgusIndexRecord (a1,  a1farhdr);
      a2DSRStatus = ArgusIndexRecord (tmp, a2farhdr);

      status = a1DSRStatus & a2DSRStatus;

      far1 = (struct ArgusFarStruct *) a1farhdr[ARGUS_FAR_DSR_INDEX];
      far2 = (struct ArgusFarStruct *) a2farhdr[ARGUS_FAR_DSR_INDEX];

      if (!(RaThisFlowRev)) {
         if (((far2->src.count + far1->src.count) < far2->src.count) ||
             ((far2->dst.count + far1->dst.count) < far2->dst.count) ||
             ((far2->src.bytes + far1->src.bytes) < far2->src.bytes) ||
             ((far2->dst.bytes + far1->dst.bytes) < far2->dst.bytes)) {
            goto doneRaMergeArgus;
         }
 
      } else {
         if (((far2->src.count + far1->dst.count) < far2->src.count) ||
             ((far2->dst.count + far1->src.count) < far2->dst.count) ||
             ((far2->src.bytes + far1->dst.bytes) < far2->src.bytes) ||
             ((far2->dst.bytes + far1->src.bytes) < far2->dst.bytes)) {
            goto doneRaMergeArgus;
         }

         if ((far1->time.start.tv_sec <  far2->time.start.tv_sec) ||
            ((far1->time.start.tv_sec == far2->time.start.tv_sec) &&
             (far1->time.start.tv_usec < far2->time.start.tv_usec))) {
            rev++;
         }
      }

      for (i = 0; i < 32; i++) {
         if (status & 0x01) {
            switch (i) {
               case ARGUS_FAR_DSR_INDEX: {
                  u_int diffstartime = 0;
                  int diffusec = 0;

                  far1 = (struct ArgusFarStruct *) a1farhdr[i];
                  far2 = (struct ArgusFarStruct *) a2farhdr[i];

                  switch (a2->ahdr.status & 0xFFFF) {
                     case ETHERTYPE_IP: {
                        if (!(RaThisFlowRev)) {
                           if (!(data->status & ARGUS_FAR_SRCADDR_MODIFIED))
                              if (RaArgusFlow->ip_flow.ip_src != data->flow.ip_flow.ip_src) {
                                 data->status |= ARGUS_FAR_SRCADDR_MODIFIED;
                                 if (far2->flow.ip_flow.ip_src != far1->flow.ip_flow.ip_src)
                                    far2->flow.ip_flow.ip_src = far1->flow.ip_flow.ip_src;
                              }
                           
                           if (!(data->status & ARGUS_FAR_DSTADDR_MODIFIED))
                              if (RaArgusFlow->ip_flow.ip_dst != data->flow.ip_flow.ip_dst) {
                                 data->status |= ARGUS_FAR_DSTADDR_MODIFIED;
                                 if (far2->flow.ip_flow.ip_dst != far1->flow.ip_flow.ip_dst)
                                    far2->flow.ip_flow.ip_dst = far1->flow.ip_flow.ip_dst;
                              }
       
                           if (!(data->status & ARGUS_FAR_PROTO_MODIFIED))
                              if (RaArgusFlow->ip_flow.ip_p != data->flow.ip_flow.ip_p) {
                                 data->status |= ARGUS_FAR_PROTO_MODIFIED;
                                 if (far2->flow.ip_flow.ip_p != far1->flow.ip_flow.ip_p)
                                    far2->flow.ip_flow.ip_p = far1->flow.ip_flow.ip_p;
                              }
       
                           switch (data->flow.ip_flow.ip_p) {
                              case IPPROTO_UDP:              
                              case IPPROTO_TCP:              
                                 far2->flow.ip_flow.ip_id = far1->flow.ip_flow.ip_id;

                                 if (!(data->status & ARGUS_FAR_SRCPORT_MODIFIED))
                                    if (RaArgusFlow->ip_flow.sport != data->flow.ip_flow.sport) {
                                       data->status |= ARGUS_FAR_SRCPORT_MODIFIED;
                                       if (far2->flow.ip_flow.sport != far1->flow.ip_flow.sport)
                                          far2->flow.ip_flow.sport = 0xFFFF;
                                    }
       
                                 if (!(data->status & ARGUS_FAR_DSTPORT_MODIFIED))
                                    if (RaArgusFlow->ip_flow.dport != data->flow.ip_flow.dport) {
                                       data->status |= ARGUS_FAR_DSTPORT_MODIFIED;
                                       if (far2->flow.ip_flow.dport != far1->flow.ip_flow.dport)
                                          far2->flow.ip_flow.dport = 0xFFFF;
                                    }
                                 break;

                              case IPPROTO_IGMP:
                                 break;

                              case IPPROTO_ICMP:
                                 if (far2->flow.icmp_flow.type != far1->flow.icmp_flow.type)
                                    far2->flow.icmp_flow.type = 0xFF;
                                 if (far2->flow.icmp_flow.code != far1->flow.icmp_flow.code)
                                    far2->flow.icmp_flow.code = 0xFF;

                                 far2->flow.icmp_flow.ip_id = far1->flow.icmp_flow.ip_id;
                                 break;
                           }
       
                           if (!(data->status & ARGUS_FAR_TPVAL_MODIFIED))
                              if (RaArgusFlow->ip_flow.tp_p != data->flow.ip_flow.tp_p) {
                                 data->status |= ARGUS_FAR_TPVAL_MODIFIED;
                                 if (far2->flow.ip_flow.tp_p != far1->flow.ip_flow.tp_p)
                                    far2->flow.ip_flow.tp_p = 0;
                              }

                        } else {
                           if (!(data->status & ARGUS_FAR_SRCADDR_MODIFIED))
                              if (RaArgusFlow->ip_flow.ip_src != data->flow.ip_flow.ip_src) {
                                 data->status |= ARGUS_FAR_SRCADDR_MODIFIED;
                                 if (far2->flow.ip_flow.ip_src != far1->flow.ip_flow.ip_dst)
                                    far2->flow.ip_flow.ip_src = far1->flow.ip_flow.ip_src;
                              }
                           
                           if (!(data->status & ARGUS_FAR_DSTADDR_MODIFIED))
                              if (RaArgusFlow->ip_flow.ip_dst != data->flow.ip_flow.ip_dst) {
                                 data->status |= ARGUS_FAR_DSTADDR_MODIFIED;
                                 if (far2->flow.ip_flow.ip_dst != far1->flow.ip_flow.ip_src)
                                    far2->flow.ip_flow.ip_dst = far1->flow.ip_flow.ip_dst;
                              }
       
                           if (!(data->status & ARGUS_FAR_PROTO_MODIFIED))
                              if (RaArgusFlow->ip_flow.ip_p != data->flow.ip_flow.ip_p) {
                                 data->status |= ARGUS_FAR_PROTO_MODIFIED;
                                 if (far2->flow.ip_flow.ip_p != far1->flow.ip_flow.ip_p)
                                    far2->flow.ip_flow.ip_p = far1->flow.ip_flow.ip_p;
                              }
       
                           switch (far2->flow.ip_flow.ip_p) {
                              case IPPROTO_UDP:
                              case IPPROTO_TCP:
                                 if (!(data->status & ARGUS_FAR_SRCPORT_MODIFIED))
                                    if (RaArgusFlow->ip_flow.sport != data->flow.ip_flow.sport) {
                                       data->status |= ARGUS_FAR_SRCPORT_MODIFIED;
                                       if (far2->flow.ip_flow.sport != far1->flow.ip_flow.dport)
                                          far2->flow.ip_flow.sport = 0xFFFF;
                                    }
       
                                 if (!(data->status & ARGUS_FAR_DSTPORT_MODIFIED))
                                    if (RaArgusFlow->ip_flow.dport != data->flow.ip_flow.dport) {
                                       data->status |= ARGUS_FAR_DSTPORT_MODIFIED;
                                       if (far2->flow.ip_flow.dport != far1->flow.ip_flow.sport)
                                          far2->flow.ip_flow.dport = 0xFFFF;
                                    }
                                 break;

                              case IPPROTO_IGMP:
                                 break;

                              case IPPROTO_ICMP:
                                 break;
                           }

                           if (!(data->status & ARGUS_FAR_TPVAL_MODIFIED))
                              if (RaArgusFlow->ip_flow.tp_p != data->flow.ip_flow.tp_p) {
                                 data->status |= ARGUS_FAR_TPVAL_MODIFIED;
                                 if (far2->flow.ip_flow.tp_p != far1->flow.ip_flow.tp_p)
                                    far2->flow.ip_flow.tp_p = 0;
                              }

                        }
                     }
                  }

                  if (!(RaThisFlowRev)) {
                     if (((far2->src.count + far1->src.count) < far2->src.count) ||
                         ((far2->src.bytes + far1->src.bytes) < far2->src.bytes) ||
                         ((far2->dst.count + far1->dst.count) < far2->dst.count) ||
                         ((far2->dst.bytes + far1->dst.bytes) < far2->dst.bytes))
                        RaSendArgusRecord(store);

                     far2->src.count    += far1->src.count;
                     far2->src.bytes    += far1->src.bytes;
                     far2->dst.count    += far1->dst.count;
                     far2->dst.bytes    += far1->dst.bytes;
                     far2->src.appbytes += far1->src.appbytes;
                     far2->dst.appbytes += far1->dst.appbytes;

                     if ((a1->ahdr.status & 0xFFFF) == ETHERTYPE_IP) {
                        if (far1->attr_ip.sttl != far2->attr_ip.sttl) far2->attr_ip.sttl = 0xFF;
                        if (far1->attr_ip.dttl != far2->attr_ip.dttl) far2->attr_ip.dttl = 0xFF;
                        if (far1->attr_ip.stos != far2->attr_ip.stos) far2->attr_ip.stos = 0xFF;
                        if (far1->attr_ip.dtos != far2->attr_ip.dtos) far2->attr_ip.dtos = 0xFF;
                     }

                  } else {
                     if (((far2->src.count + far1->dst.count) < far2->src.count) ||
                         ((far2->src.bytes + far1->dst.bytes) < far2->src.bytes) ||
                         ((far2->dst.count + far1->src.count) < far2->dst.count) ||
                         ((far2->dst.bytes + far1->src.bytes) < far2->dst.bytes))
                        RaSendArgusRecord(store);

                     if ((a1->ahdr.status & 0xFFFF) == ETHERTYPE_IP) {
                        if (far1->attr_ip.sttl != far2->attr_ip.dttl) far2->attr_ip.dttl = 0xFF;
                        if (far1->attr_ip.dttl != far2->attr_ip.sttl) far2->attr_ip.sttl = 0xFF;
                        if (far1->attr_ip.stos != far2->attr_ip.dtos) far2->attr_ip.dtos = 0xFF;
                        if (far1->attr_ip.dtos != far2->attr_ip.stos) far2->attr_ip.stos = 0xFF;
                     }

                     far2->src.count    += far1->dst.count;
                     far2->src.bytes    += far1->dst.bytes;
                     far2->dst.count    += far1->src.count;
                     far2->dst.bytes    += far1->src.bytes;
                     far2->src.appbytes += far1->dst.appbytes;
                     far2->dst.appbytes += far1->src.appbytes;
                  }

                  diffstartime  = (far1->time.start.tv_sec  - data->agr.laststartime.tv_sec);
                  if ((diffusec = (far1->time.start.tv_usec - data->agr.laststartime.tv_usec)) < 0) {
                     diffstartime--;
                     diffusec += 1000000;
                  }

                  data->agr.laststartime = far1->time.start;

                  if (diffstartime > 0) {
                     if (diffstartime > 2147) {
                        if (data->agr.status & ARGUS_AGR_USECIDLETIME) {
                           if (data->agr.idle.minval != 0x7FFFFFFF)
                              data->agr.idle.minval /= 1000;
                           data->agr.idle.maxval /= 1000;
                           data->idle.sumtime /= 1000;
                           data->idle.sumsqrd /= 1000000.0;
                           data->agr.status &= ~ARGUS_AGR_USECIDLETIME;
                           data->agr.status |=  ARGUS_AGR_MSECIDLETIME;
                        }
                     }

                     if (data->agr.status & ARGUS_AGR_USECIDLETIME)
                        TimeMultiplier = 1000000;
                     else
                        TimeMultiplier = 1000;

                     diffstartime *= TimeMultiplier;
                     diffstartime += (diffusec * 1000) / (TimeMultiplier / 1000);

                     if ((data->agr.idle.minval > diffstartime) && (diffstartime > 0))
                        data->agr.idle.minval = diffstartime;

                     if ((data->agr.idle.maxval < diffstartime) && (diffstartime > 0))
                        data->agr.idle.maxval = diffstartime;

                     data->idle.sumtime += diffstartime;
                     data->idle.sumsqrd += pow(diffstartime, 2.0);
                     data->idle.n++;
                  }

                  duration = RaGetuSecDuration(a1);
                  if (duration > 0x6FFFFFFF) {
                     if (data->agr.status & ARGUS_AGR_USECIDLETIME) {
                         if (data->agr.idle.minval != 0x7FFFFFFF)
                            data->agr.act.minval /= 1000;
                         data->agr.act.maxval /= 1000;
                         data->act.sumtime /= 1000;
                         data->act.sumsqrd /= 1000000.0;
                         data->agr.status &= ~ARGUS_AGR_USECACTTIME;
                         data->agr.status |=  ARGUS_AGR_MSECACTTIME;
                     }
                  }

                  if (data->agr.status & ARGUS_AGR_USECACTTIME)
                     TimeMultiplier = 1;
                  else
                     TimeMultiplier = 1000;

                  duration /= TimeMultiplier;

                  if ((data->agr.act.minval > duration) && (duration > 0))
                     data->agr.act.minval = duration;

                  if ((data->agr.act.maxval < duration) && (duration > 0))
                     data->agr.act.maxval = duration;

                  data->act.sumtime += duration;
                  data->act.sumsqrd += pow(duration, 2.0);
                  data->act.n++;

                  if (((a1->ahdr.status & 0xFFFF) == ETHERTYPE_IP) && (far2->flow.ip_flow.ip_p == IPPROTO_TCP)) {
                     struct ArgusTCPObject *tcp1 = (struct ArgusTCPObject *) a1farhdr[ARGUS_TCP_DSR_INDEX];

                     if (tcp1 && ((tcp1->state & (ARGUS_FIN | ARGUS_FIN_ACK | ARGUS_NORMAL_CLOSE | ARGUS_RESET)) > 0)) {
                        if (!((tcp1->state & ARGUS_RESET) && (!(far1->dst.count && far1->src.count))))
                           data->agr.count++;
                     }
 
                  } else
                     data->agr.count++;

                  if ((agr = (struct ArgusAGRStruct *) a2farhdr[ARGUS_AGR_DSR_INDEX]) == NULL) {
                  }

                  if (far1->ArgusTransRefNum != far2->ArgusTransRefNum)
                     far2->ArgusTransRefNum = 0;

                  if ((far2->time.start.tv_sec  > far1->time.start.tv_sec) ||
                     ((far2->time.start.tv_sec == far1->time.start.tv_sec) &&
                     (far2->time.start.tv_usec  > far1->time.start.tv_usec))) {

                     far2->time.start = far1->time.start;
                     if (RaThisFlowRev)
                        rev++;
                  }

                  if ((far2->time.last.tv_sec  < far1->time.last.tv_sec) ||
                     ((far2->time.last.tv_sec == far1->time.last.tv_sec) &&
                     (far2->time.last.tv_usec  < far1->time.last.tv_usec)))

                     far2->time.last = far1->time.last;

                  if (store->qhdr.lasttime.tv_sec != far2->time.last.tv_sec) {
                     store->qhdr.lasttime = far2->time.last;
                     data->agr.lasttime = far2->time.last;
                  } else
                     if (store->qhdr.lasttime.tv_usec < far2->time.last.tv_usec) {
                        store->qhdr.lasttime = far2->time.last;
                     }
               }
               break;

#define TCP_FIN_STATE_TIME	360

               case ARGUS_TCP_DSR_INDEX: {
                  struct ArgusTCPObject *tcp1 = (struct ArgusTCPObject *) a1farhdr[i];
                  struct ArgusTCPObject *tcp2 = (struct ArgusTCPObject *) a2farhdr[i];

                  if ((tcp1->state & ARGUS_SAW_SYN) && (!(tcp2->state & ARGUS_SAW_SYN)))
                     if (RaThisFlowRev)
                        rev++;

                  if (RaCumulativeMerge)
                     tcp2->state |= tcp1->state;
                  else
                     tcp2->state = tcp1->state;

                  if (!(tcp2->synAckuSecs))
                     tcp2->synAckuSecs = tcp1->synAckuSecs;
                  else
                     if (tcp2->synAckuSecs > tcp1->synAckuSecs)
                        if (tcp1->synAckuSecs)
                           tcp2->synAckuSecs = tcp1->synAckuSecs;

                  if (!(tcp2->ackDatauSecs))
                     tcp2->ackDatauSecs = tcp1->ackDatauSecs;
                  else
                     if (tcp2->ackDatauSecs > tcp1->ackDatauSecs)
                        if (tcp1->ackDatauSecs)
                           tcp2->ackDatauSecs = tcp1->ackDatauSecs;

                  tcp2->src.seqbase   = 0;
                  tcp2->dst.seqbase   = 0;

                  if (!(RaThisFlowRev)) {
                     tcp2->src.ackbytes += tcp1->src.ackbytes;
                     tcp2->dst.ackbytes += tcp1->dst.ackbytes;
                     tcp2->src.bytes    += tcp1->src.bytes;
                     tcp2->dst.bytes    += tcp1->dst.bytes;
                     tcp2->src.rpkts    += tcp1->src.rpkts;
                     tcp2->dst.rpkts    += tcp1->dst.rpkts;        
 
                     if (tcp2->src.win > tcp1->src.win);
                        tcp2->src.win = tcp1->src.win;
                     if (tcp2->dst.win > tcp1->dst.win);
                        tcp2->dst.win = tcp1->dst.win;        

                     tcp2->src.flags    |= tcp1->src.flags;
                     tcp2->dst.flags    |= tcp1->dst.flags;        

                  } else {
                     tcp2->src.ackbytes += tcp1->dst.ackbytes;
                     tcp2->dst.ackbytes += tcp1->src.ackbytes;
                     tcp2->src.bytes    += tcp1->dst.bytes;
                     tcp2->dst.bytes    += tcp1->src.bytes;
                     tcp2->src.rpkts    += tcp1->dst.rpkts;
                     tcp2->dst.rpkts    += tcp1->src.rpkts;        
 
                     if (tcp2->src.win > tcp1->dst.win);
                        tcp2->src.win = tcp1->dst.win;
                     if (tcp2->dst.win > tcp1->src.win);
                        tcp2->dst.win = tcp1->src.win;        

                     tcp2->src.flags    |= tcp1->dst.flags;
                     tcp2->dst.flags    |= tcp1->src.flags;        
                  }
               }
               break;

               case ARGUS_RTP_DSR_INDEX: {
                  struct ArgusRTPObject *rtp1 = (struct ArgusRTPObject *) a1farhdr[i];
                  struct ArgusRTPObject *rtp2 = (struct ArgusRTPObject *) a2farhdr[i];

                  if (rtp1 && rtp2) {
                     if (!(RaThisFlowRev)) {
                        if ((((rtp2->sdrop + rtp1->sdrop) < rtp2->sdrop) || ((rtp2->ddrop + rtp1->ddrop) < rtp2->ddrop)) ||
                            (((rtp2->ssdev + rtp1->ssdev) < rtp2->ssdev) || ((rtp2->dsdev + rtp1->dsdev) < rtp2->dsdev))) {
                           RaSendArgusRecord(store);
                           rtp2->sdrop = rtp1->sdrop;
                           rtp2->ddrop = rtp1->ddrop;
                           rtp2->ssdev = rtp1->ssdev;
                           rtp2->dsdev = rtp1->dsdev;
 
                        } else {
                           rtp2->sdrop += rtp1->sdrop;
                           rtp2->ddrop += rtp1->ddrop;
                           rtp2->ssdev += rtp1->ssdev;
                           rtp2->dsdev += rtp1->dsdev;
                        }

                     } else {
                        if ((((rtp2->ddrop + rtp1->sdrop) < rtp2->ddrop) || ((rtp2->sdrop + rtp1->ddrop) < rtp2->sdrop)) ||
                            (((rtp2->dsdev + rtp1->ssdev) < rtp2->dsdev) || ((rtp2->ssdev + rtp1->dsdev) < rtp2->ssdev))) {
                           RaSendArgusRecord(store);
                           rtp2->sdrop = rtp1->ddrop;
                           rtp2->ddrop = rtp1->sdrop;
                           rtp2->ssdev = rtp1->dsdev;
                           rtp2->dsdev = rtp1->ssdev;
 
                        } else {
                           rtp2->sdrop += rtp1->ddrop;
                           rtp2->ddrop += rtp1->sdrop;
                           rtp2->ssdev += rtp1->dsdev;
                           rtp2->dsdev += rtp1->ssdev;
                        }
                     }
                  }
               }
               break;

               case ARGUS_ICMP_DSR_INDEX: {
                  struct ArgusICMPObject *icmp1 = (struct ArgusICMPObject *) a1farhdr[i]; 
                  struct ArgusICMPObject *icmp2 = (struct ArgusICMPObject *) a2farhdr[i]; 

                  if (icmp1->icmp_type != icmp2->icmp_type) {
                     switch (icmp2->icmp_type) {
                        case ICMP_MASKREQ:
                        case ICMP_ECHO:
                        case ICMP_TSTAMP:
                        case ICMP_IREQ:
                           break;

                        case ICMP_MASKREPLY:
                           if (icmp1->icmp_type == ICMP_MASKREQ) {
                              icmp2->icmp_type = ICMP_MASKREQ;
                              rev++;
                           }
                           break;
                        case ICMP_ECHOREPLY:
                           if (icmp1->icmp_type == ICMP_ECHO) {
                              icmp2->icmp_type = ICMP_ECHO;
                              rev++;
                           }
                           break;
                        case ICMP_TSTAMPREPLY:
                           if (icmp1->icmp_type == ICMP_TSTAMP) {
                              icmp2->icmp_type = ICMP_TSTAMP;
                              rev++;
                           }
                           break;
                        case ICMP_IREQREPLY:
                           if (icmp1->icmp_type == ICMP_IREQ) {
                              icmp2->icmp_type = ICMP_IREQ;
                              rev++;
                           }
                           break;
                     }
                  }
               }


               case ARGUS_ARP_DSR_INDEX:
               case ARGUS_FRG_DSR_INDEX:
                  break;

               case ARGUS_VLAN_DSR_INDEX: {
                  struct ArgusVlanStruct *vlan1 = (struct ArgusVlanStruct *) a1farhdr[i];
                  struct ArgusVlanStruct *vlan2 = (struct ArgusVlanStruct *) a2farhdr[i];

                  if (vlan1 && vlan2) {
                     if ((vlan2->status & ARGUS_SRC_VLAN) && (vlan1->status & ARGUS_SRC_VLAN))
                        if (vlan2->sid != vlan1->sid)
                           vlan2->status |= ARGUS_SRC_CHANGED;

                     if ((vlan2->status & ARGUS_DST_VLAN) && (vlan1->status & ARGUS_DST_VLAN))
                        if (vlan2->did != vlan1->did)
                           vlan2->status |= ARGUS_DST_CHANGED;
                  
                     vlan2->status |= vlan1->status;
                     vlan2->sid = vlan1->sid;
                     vlan2->did = vlan1->did;
                  }
               }
               break;

               case ARGUS_MPLS_DSR_INDEX: {
                  struct ArgusMplsStruct *mpls1 = (struct ArgusMplsStruct *) a1farhdr[i];
                  struct ArgusMplsStruct *mpls2 = (struct ArgusMplsStruct *) a2farhdr[i];

                  if (mpls1 && mpls2) {
                     if (!(RaThisFlowRev)) {
                        if ((mpls2->status & ARGUS_SRC_MPLS) && (mpls1->status & ARGUS_SRC_MPLS))
                           if (mpls2->slabel != mpls1->slabel)
                              mpls2->status |= ARGUS_SRC_CHANGED;
 
                        if ((mpls2->status & ARGUS_DST_MPLS) && (mpls1->status & ARGUS_DST_MPLS))
                           if (mpls2->dlabel != mpls1->dlabel)
                              mpls2->status |= ARGUS_DST_CHANGED;
                     
                        mpls2->status |= mpls1->status;
                        mpls2->slabel = mpls1->slabel;
                        mpls2->dlabel = mpls1->dlabel;

                     } else {
                        if ((mpls2->status & ARGUS_SRC_MPLS) && (mpls1->status & ARGUS_SRC_MPLS))
                           if (mpls2->slabel != mpls1->dlabel)
                              mpls2->status |= ARGUS_SRC_CHANGED;
 
                        if ((mpls2->status & ARGUS_DST_MPLS) && (mpls1->status & ARGUS_DST_MPLS))
                           if (mpls2->dlabel != mpls1->slabel)
                              mpls2->status |= ARGUS_DST_CHANGED;
                     
                        mpls2->status |= mpls1->status;
                        mpls2->slabel = mpls1->dlabel;
                        mpls2->dlabel = mpls1->slabel;
                     }
                  }
               }
               break;

               case ARGUS_ESP_DSR_INDEX: {
                  struct ArgusESPStruct *esp1 = (struct ArgusESPStruct *) a1farhdr[i];
                  struct ArgusESPStruct *esp2 = (struct ArgusESPStruct *) a2farhdr[i];

                  if (esp1 && esp2) {
                     if (!(RaThisFlowRev)) {
                        if (esp1->src.spi != esp2->src.spi)
                           esp2->src.spi = 0;
 
                        if (esp1->dst.spi != esp2->dst.spi)
                           esp2->dst.spi = 0;
 
                        esp2->src.lastseq = esp1->src.lastseq;
                        esp2->dst.lastseq = esp1->dst.lastseq;
                        esp2->src.lostseq = esp1->src.lostseq;
                        esp2->dst.lostseq = esp1->dst.lostseq;

                     } else {
                        if (esp1->src.spi != esp2->dst.spi)
                           esp2->dst.spi = 0;
 
                        if (esp1->dst.spi != esp2->src.spi)
                           esp2->src.spi = 0;
 
                        esp2->src.lastseq = esp1->dst.lastseq;
                        esp2->dst.lastseq = esp1->src.lastseq;
                        esp2->src.lostseq = esp1->dst.lostseq;
                        esp2->dst.lostseq = esp1->src.lostseq;
                     }
                  }
               }
               break;
    
               case ARGUS_TIME_DSR_INDEX: {
                  struct ArgusTimeStruct *time1 = (struct ArgusTimeStruct *) a1farhdr[i];
                  struct ArgusTimeStruct *time2 = (struct ArgusTimeStruct *) a2farhdr[i];
                  long long sumtime, sumtime1, sumtime2;
                  double sumsqrd, sumsqrd1, sumsqrd2;

                  if (time1 && time2) {
                     if (!(RaThisFlowRev)) {
                        if (time1->src.act.n > 0) {
                           sumtime1 = (time1->src.act.meanval * time1->src.act.n);
                           sumtime2 = (time2->src.act.meanval * time2->src.act.n);
                           if ((sumtime = (sumtime1 + sumtime2)) < sumtime1)
                              goto doneRaMergeArgus;
   
                           if (time1->src.act.n)
                              sumsqrd1 = (time1->src.act.n * pow(time1->src.act.stdev, 2.0)) + pow(sumtime1, 2.0)/time1->src.act.n;
                           else
                              sumsqrd1 = 0;
     
                           if (time2->src.act.n)
                              sumsqrd2 = (time2->src.act.n * pow(time2->src.act.stdev, 2.0)) + pow(sumtime2, 2.0)/time2->src.act.n;
                           else
                              sumsqrd2 = 0;
   
                           time2->src.act.n  += time1->src.act.n;
                           time2->src.act.meanval  = sumtime/time2->src.act.n;
   
                           if ((sumsqrd = sumsqrd1 + sumsqrd2) < sumsqrd1)
                              goto doneRaMergeArgus;
     
                           time2->src.act.stdev = sqrt (sumsqrd/time2->src.act.n - pow (sumtime/time2->src.act.n, 2.0));
                           if (time2->src.act.maxval < time1->src.act.maxval)
                              time2->src.act.maxval = time1->src.act.maxval; 
                           if (time2->src.act.minval > time1->src.act.minval)
                              time2->src.act.minval = time1->src.act.minval; 
                        }
   
                        if (time1->dst.act.n > 0) {
                           sumtime1 = (time1->dst.act.meanval * time1->dst.act.n);
                           sumtime2 = (time2->dst.act.meanval * time2->dst.act.n);
                           if ((sumtime = sumtime1 + sumtime2) < sumtime1)
                              goto doneRaMergeArgus;
     
                           if (time1->dst.act.n)
                              sumsqrd1 = (time1->dst.act.n * pow(time1->dst.act.stdev, 2.0)) + pow(sumtime1, 2.0)/time1->dst.act.n;
                           else
                              sumsqrd1 = 0;
     
                           sumsqrd2 = (time2->dst.act.n * pow(time2->dst.act.stdev, 2.0)) + pow(sumtime2, 2.0)/time2->dst.act.n;
                           if ((sumsqrd = sumsqrd1 + sumsqrd2) < sumsqrd1)
                              goto doneRaMergeArgus;
     
                           time2->dst.act.n  += time1->dst.act.n;
                           time2->dst.act.meanval  = sumtime/time2->dst.act.n;
   
                           time2->dst.act.stdev = sqrt (sumsqrd/time2->dst.act.n - pow (sumtime/time2->dst.act.n, 2.0));
                           if (time2->dst.act.maxval < time1->dst.act.maxval)
                              time2->dst.act.maxval = time1->dst.act.maxval; 
                           if (time2->dst.act.minval > time1->dst.act.minval)
                              time2->dst.act.minval = time1->dst.act.minval; 
                        }
   
                        if (time1->src.idle.n > 0) {
                           sumtime1 = (time1->src.idle.meanval * time1->src.idle.n);
                           sumtime2 = (time2->src.idle.meanval * time2->src.idle.n);
                           if ((sumtime = sumtime1 + sumtime2) < sumtime1)
                              goto doneRaMergeArgus;
     
                           time2->src.idle.n  += time1->src.idle.n;
                           time2->src.idle.meanval  = sumtime/time2->src.idle.n;
                           if (time1->src.idle.n)
                              sumsqrd1 = (time1->src.idle.n * pow(time1->src.idle.stdev, 2.0)) + pow(sumtime1, 2.0)/time1->src.idle.n;
                           else
                              sumsqrd1 = 0;
     
                           sumsqrd2 = (time2->src.idle.n * pow(time2->src.idle.stdev, 2.0)) + pow(sumtime2, 2.0)/time2->src.idle.n;
                           if ((sumsqrd = sumsqrd1 + sumsqrd2) < sumsqrd1)
                              goto doneRaMergeArgus;
     
                           time2->src.idle.stdev = sqrt (sumsqrd/time2->src.idle.n - pow (sumtime/time2->src.idle.n, 2.0));
                           if (time2->src.idle.maxval < time1->src.idle.maxval)
                              time2->src.idle.maxval = time1->src.idle.maxval; 
                           if (time2->src.idle.minval > time1->src.idle.minval)
                              time2->src.idle.minval = time1->src.idle.minval; 
                        }
   
                        if (time1->dst.idle.n > 0) {
                           sumtime1 = (time1->dst.idle.meanval * time1->dst.idle.n);
                           sumtime2 = (time2->dst.idle.meanval * time2->dst.idle.n);
                           if ((sumtime = sumtime1 + sumtime2) < sumtime1)
                              goto doneRaMergeArgus;
     
                           time2->dst.idle.n  += time1->dst.idle.n;
                           time2->dst.idle.meanval  = sumtime/time2->dst.idle.n;
                           if (time1->dst.idle.n)
                              sumsqrd1 = (time1->dst.idle.n * pow(time1->dst.idle.stdev, 2.0)) + pow(sumtime1, 2.0)/time1->dst.idle.n;
                           else
                              sumsqrd1 = 0;
     
                           sumsqrd2 = (time2->dst.idle.n * pow(time2->dst.idle.stdev, 2.0)) + pow(sumtime2, 2.0)/time2->dst.idle.n;
                           if ((sumsqrd = sumsqrd1 + sumsqrd2) < sumsqrd1)
                              goto doneRaMergeArgus;
     
                           time2->dst.idle.stdev = sqrt (sumsqrd/time2->dst.idle.n - pow (sumtime/time2->dst.idle.n, 2.0));
                           if (time2->dst.idle.maxval < time1->dst.idle.maxval)
                              time2->dst.idle.maxval = time1->dst.idle.maxval; 
                           if (time2->dst.idle.minval > time1->dst.idle.minval)
                              time2->dst.idle.minval = time1->dst.idle.minval; 
                        }

                     } else {
                        if (time1->src.act.n > 0) {
                           sumtime1 = (time1->src.act.meanval * time1->src.act.n);
                           sumtime2 = (time2->dst.act.meanval * time2->dst.act.n);
                           if ((sumtime = (sumtime1 + sumtime2)) < sumtime1)
                              goto doneRaMergeArgus;
   
                           if (time1->src.act.n)
                              sumsqrd1 = (time1->src.act.n * pow(time1->src.act.stdev, 2.0)) + pow(sumtime1, 2.0)/time1->src.act.n;
                           else
                              sumsqrd1 = 0;
     
                           if (time2->dst.act.n)
                              sumsqrd2 = (time2->dst.act.n * pow(time2->dst.act.stdev, 2.0)) + pow(sumtime2, 2.0)/time2->dst.act.n;
                           else
                              sumsqrd2 = 0;
   
                           time2->dst.act.n  += time1->src.act.n;
                           time2->dst.act.meanval  = sumtime/time2->dst.act.n;
   
                           if ((sumsqrd = sumsqrd1 + sumsqrd2) < sumsqrd1)
                              goto doneRaMergeArgus;
     
                           time2->dst.act.stdev = sqrt (sumsqrd/time2->dst.act.n - pow (sumtime/time2->dst.act.n, 2.0));
                           if (time2->dst.act.maxval < time1->src.act.maxval)
                              time2->dst.act.maxval = time1->src.act.maxval; 
                           if (time2->dst.act.minval > time1->src.act.minval)
                              time2->dst.act.minval = time1->src.act.minval; 
                        }
   
                        if (time1->dst.act.n > 0) {
                           sumtime1 = (time1->dst.act.meanval * time1->dst.act.n);
                           sumtime2 = (time2->src.act.meanval * time2->src.act.n);
                           if ((sumtime = sumtime1 + sumtime2) < sumtime1)
                              goto doneRaMergeArgus;
     
                           if (time1->dst.act.n)
                              sumsqrd1 = (time1->dst.act.n * pow(time1->dst.act.stdev, 2.0)) + pow(sumtime1, 2.0)/time1->dst.act.n;
                           else
                              sumsqrd1 = 0;
     
                           sumsqrd2 = (time2->src.act.n * pow(time2->src.act.stdev, 2.0)) + pow(sumtime2, 2.0)/time2->src.act.n;
                           if ((sumsqrd = sumsqrd1 + sumsqrd2) < sumsqrd1)
                              goto doneRaMergeArgus;
     
                           time2->src.act.n  += time1->dst.act.n;
                           time2->src.act.meanval  = sumtime/time2->src.act.n;
   
                           time2->src.act.stdev = sqrt (sumsqrd/time2->src.act.n - pow (sumtime/time2->src.act.n, 2.0));
                           if (time2->src.act.maxval < time1->dst.act.maxval)
                              time2->src.act.maxval = time1->dst.act.maxval; 
                           if (time2->src.act.minval > time1->dst.act.minval)
                              time2->src.act.minval = time1->dst.act.minval; 
                        }
   
                        if (time1->src.idle.n > 0) {
                           sumtime1 = (time1->src.idle.meanval * time1->src.idle.n);
                           sumtime2 = (time2->dst.idle.meanval * time2->dst.idle.n);
                           if ((sumtime = sumtime1 + sumtime2) < sumtime1)
                              goto doneRaMergeArgus;
     
                           time2->dst.idle.n  += time1->src.idle.n;
                           time2->dst.idle.meanval  = sumtime/time2->dst.idle.n;
                           if (time1->src.idle.n)
                              sumsqrd1 = (time1->src.idle.n * pow(time1->src.idle.stdev, 2.0)) + pow(sumtime1, 2.0)/time1->src.idle.n;
                           else
                              sumsqrd1 = 0;
     
                           sumsqrd2 = (time2->dst.idle.n * pow(time2->dst.idle.stdev, 2.0)) + pow(sumtime2, 2.0)/time2->dst.idle.n;
                           if ((sumsqrd = sumsqrd1 + sumsqrd2) < sumsqrd1)
                              goto doneRaMergeArgus;
     
                           time2->dst.idle.stdev = sqrt (sumsqrd/time2->dst.idle.n - pow (sumtime/time2->dst.idle.n, 2.0));
                           if (time2->dst.idle.maxval < time1->src.idle.maxval)
                              time2->dst.idle.maxval = time1->src.idle.maxval; 
                           if (time2->dst.idle.minval > time1->src.idle.minval)
                              time2->dst.idle.minval = time1->src.idle.minval; 
                        }
   
                        if (time1->dst.idle.n > 0) {
                           sumtime1 = (time1->dst.idle.meanval * time1->dst.idle.n);
                           sumtime2 = (time2->src.idle.meanval * time2->src.idle.n);
                           if ((sumtime = sumtime1 + sumtime2) < sumtime1)
                              goto doneRaMergeArgus;
     
                           time2->src.idle.n  += time1->dst.idle.n;
                           time2->src.idle.meanval  = sumtime/time2->src.idle.n;
                           if (time1->dst.idle.n)
                              sumsqrd1 = (time1->dst.idle.n * pow(time1->dst.idle.stdev, 2.0)) + pow(sumtime1, 2.0)/time1->dst.idle.n;
                           else
                              sumsqrd1 = 0;
     
                           sumsqrd2 = (time2->src.idle.n * pow(time2->src.idle.stdev, 2.0)) + pow(sumtime2, 2.0)/time2->src.idle.n;
                           if ((sumsqrd = sumsqrd1 + sumsqrd2) < sumsqrd1)
                              goto doneRaMergeArgus;
     
                           time2->src.idle.stdev = sqrt (sumsqrd/time2->src.idle.n - pow (sumtime/time2->src.idle.n, 2.0));
                           if (time2->src.idle.maxval < time1->dst.idle.maxval)
                              time2->src.idle.maxval = time1->dst.idle.maxval; 
                           if (time2->src.idle.minval > time1->dst.idle.minval)
                              time2->src.idle.minval = time1->dst.idle.minval; 
                        }
                     }
                  }
               }
               break;

               case ARGUS_SRCUSRDATA_DSR_INDEX:
               case ARGUS_DSTUSRDATA_DSR_INDEX: {
                  struct ArgusUserStruct *user1 = (struct ArgusUserStruct *) a1farhdr[i];
                  struct ArgusUserStruct *user2 = (struct ArgusUserStruct *) a2farhdr[i];
                  int len, len1 = 0, len2 = 0, x;
                  unsigned short newstatus = 0;

                  if (user1 && user2) {
                     len1 = (user1->length - 1) * 4;
                     len2 = (user2->length - 1) * 4;

                     len = (len1 > len2) ? len2 : len1;
                     len = (len > 16) ? 16 : len;

                     for (x = 0; x < len; x++)
                        if ((&user2->data)[x] != (&user1->data)[x])
                           ((char *)&newstatus)[x/8] |= (0x80 >> (x % 8));
                     
                     user2->status = newstatus;
                  }
               }
               break;

               case ARGUS_AGR_DSR_INDEX: {
                  struct ArgusAGRStruct *agr1 = (struct ArgusAGRStruct *) a1farhdr[i];
                  unsigned int agr1sumtime, sumtime;
                  double agr1sumsqrd, sumsqrd;
                  int overlapping = 1;
                  long long newvalue = 0;

                  if (!((data->agr.status & ARGUS_AGR_USECACTTIME) &&
                            (agr1->status & ARGUS_AGR_USECACTTIME))) {

                     if (data->agr.status & ARGUS_AGR_USECACTTIME) {
                         data->agr.act.maxval   /= 1000;
                         data->agr.act.minval   /= 1000;
                         data->agr.act.meanval  /= 1000;
                         data->agr.act.stdev /= 1000;
                         data->act.sumtime /= 1000;
                         data->act.sumsqrd /= 1000000.0;
                         data->agr.status &= ~ARGUS_AGR_USECACTTIME;
                         data->agr.status |=  ARGUS_AGR_MSECACTTIME;
                     }

                     if (agr1->status & ARGUS_AGR_USECACTTIME) {
                         agr1->act.maxval   /= 1000;
                         agr1->act.minval   /= 1000;
                         agr1->act.meanval  /= 1000;
                         agr1->act.stdev /= 1000;
                         agr1->status &= ~ARGUS_AGR_USECACTTIME;
                         agr1->status |=  ARGUS_AGR_MSECACTTIME;
                     }
                  }

                  if (data->agr.act.minval > agr1->act.minval)
                     data->agr.act.minval = agr1->act.minval;
                  if (data->agr.act.maxval < agr1->act.maxval)
                     data->agr.act.maxval = agr1->act.maxval;

                  if (agr1->act.n) {
                     agr1sumtime = agr1->act.meanval * agr1->act.n;
                     agr1sumsqrd = (agr1->act.n * pow(agr1->act.stdev, 2.0)) +
                                    pow(agr1sumtime, 2.0)/agr1->act.n;

                     data->act.sumtime += agr1sumtime;
                     data->act.sumsqrd += agr1sumsqrd;

                     sumtime = data->act.sumtime;
                     sumsqrd = data->act.sumsqrd;

                     data->agr.act.n += agr1->act.n;
                     data->act.n += agr1->act.n;

                     data->agr.act.meanval = sumtime/data->act.n;
                     data->agr.act.stdev = sqrt (sumsqrd/data->act.n -
                                           pow (sumtime/data->act.n, 2.0));
                  }

                  if (!((data->agr.status & ARGUS_AGR_USECIDLETIME) &&
                            (agr1->status & ARGUS_AGR_USECIDLETIME))) {

                     if (data->agr.status & ARGUS_AGR_USECIDLETIME) {
                         data->agr.idle.maxval   /= 1000;
                         data->agr.idle.minval   /= 1000;
                         data->agr.idle.meanval  /= 1000;
                         data->agr.idle.stdev /= 1000;
                         data->idle.sumtime /= 1000;
                         data->idle.sumsqrd /= 1000000.0;
                         data->agr.status &= ~ARGUS_AGR_USECIDLETIME;
                         data->agr.status |=  ARGUS_AGR_MSECIDLETIME;
                     }

                     if (agr1->status & ARGUS_AGR_USECIDLETIME) {
                         agr1->idle.maxval   /= 1000;
                         agr1->idle.minval   /= 1000;
                         agr1->idle.meanval  /= 1000;
                         agr1->idle.stdev /= 1000;
                         agr1->status &= ~ARGUS_AGR_USECIDLETIME;
                         agr1->status |=  ARGUS_AGR_MSECIDLETIME;
                     }
                  }

                  if (data->agr.status & ARGUS_AGR_USECIDLETIME)
                     TimeMultiplier = 1;
                  else
                     TimeMultiplier = 1000;

                  overlapping = 1;

                  if ((a2->argus_far.time.start.tv_sec < a1->argus_far.time.start.tv_sec) ||
                     ((a2->argus_far.time.start.tv_sec == a1->argus_far.time.start.tv_sec) &&
                      (a2->argus_far.time.start.tv_usec < a1->argus_far.time.start.tv_usec))) {

                     if ((a2->argus_far.time.last.tv_sec < a1->argus_far.time.start.tv_sec) ||
                        ((a2->argus_far.time.last.tv_sec == a1->argus_far.time.start.tv_sec) &&
                         (a2->argus_far.time.last.tv_usec < a1->argus_far.time.start.tv_usec))) {

                        newvalue = (((agr1->laststartime.tv_sec  - data->agr.laststartime.tv_sec) * 1000000) +
                                     (agr1->laststartime.tv_usec - data->agr.laststartime.tv_usec))/TimeMultiplier;
                        overlapping = 0;
                     }

                  } else {
                     if ((a1->argus_far.time.last.tv_sec < a2->argus_far.time.start.tv_sec) ||
                        ((a1->argus_far.time.last.tv_sec == a2->argus_far.time.start.tv_sec) &&
                         (a1->argus_far.time.last.tv_usec < a2->argus_far.time.start.tv_usec))) {
                        newvalue = (((data->agr.laststartime.tv_sec  - agr1->laststartime.tv_sec) * 1000000) +
                                     (data->agr.laststartime.tv_usec - agr1->laststartime.tv_usec))/TimeMultiplier;
                        overlapping = 0;
                     }
                  }

                  if (!(overlapping)) {
                     sumtime = (agr1->idle.meanval * agr1->idle.n) + newvalue;
                     sumsqrd = ((agr1->idle.n * pow(agr1->idle.stdev, 2.0)) + pow(sumtime, 2.0)/agr1->idle.n) +
                                 pow(newvalue, 2.0);

                     data->idle.sumtime += sumtime;
                     data->idle.sumsqrd += sumsqrd;
                     data->idle.n += agr1->idle.n + 1;

                     if (newvalue) {
                        if (newvalue > data->agr.idle.maxval) 
                           data->agr.idle.maxval = newvalue;
                        else
                        if (newvalue < data->agr.idle.minval)
                           data->agr.idle.minval = newvalue;
                     }

                     data->agr.count += agr1->count;

                  } else {
                     if (data->agr.idle.minval >= agr1->idle.minval) {
                        data->idle.sumtime += agr1->idle.minval;
                        data->idle.sumsqrd += pow(agr1->idle.minval, 2.0);
                        data->idle.n++;
                        data->agr.idle.minval = agr1->idle.minval;
                     }
                  }

                  if ((data->agr.laststartime.tv_sec > agr1->laststartime.tv_sec) ||
                      ((data->agr.laststartime.tv_sec == agr1->laststartime.tv_sec) &&
                       (data->agr.laststartime.tv_usec > agr1->laststartime.tv_usec))) {
                     data->agr.laststartime = agr1->laststartime;
                  }

                  if ((data->agr.lasttime.tv_sec < agr1->lasttime.tv_sec) ||
                      ((data->agr.lasttime.tv_sec == agr1->lasttime.tv_sec) &&
                       (data->agr.lasttime.tv_usec < agr1->lasttime.tv_usec))) {

                     store->qhdr.lasttime = agr1->lasttime;
                     data->agr.lasttime = agr1->lasttime;
                  }
               }
               break;
            }

         } else {
/*
            int datalen = 0;

            switch (i) {
               case ARGUS_SRCUSRDATA_DSR_INDEX:
                  datalen = (ArgusSrcUserDataLen + 3)/ 4;

               case ARGUS_DSTUSRDATA_DSR_INDEX: {
                  struct ArgusUserStruct *user1 = (struct ArgusUserStruct *) a1farhdr[i];
                  struct ArgusUserStruct *user2 = (struct ArgusUserStruct *) a2farhdr[i];

                  datalen = (ArgusDstUserDataLen + 3)/ 4;
                  if (user1 && (user2 == NULL)) {
                     datalen = (datalen > (user1->length)) ? datalen : (user1->length);
                     user1->length = datalen + 1;
                     bcopy ((char *) user1, &buf[tmp->ahdr.length], user1->length*4);
                     tmp->ahdr.length += (user1->length * 4);
                  }

                  break;
               }
            }
*/
         }
         status >>= 1;
      }

      if (rev) {
         ArgusReverseRecord (tmp);
 
         if (store->rahtblhdr != NULL) {
            struct RaHashTableStruct *htable;

            if ((htable = store->rahtblhdr->htable) != NULL) {
               RaRemoveHashEntry (htable, store->rahtblhdr);
               RaFindArgusRecord(htable, tmp);
               if ((store->rahtblhdr = RaAddHashEntry (&RaHashTable, store)) == NULL)
                  ArgusLog (LOG_ERR, "RaMergeArgusRecord: RaAddHashEntry: %s", strerror(errno));
            }
         }
      }

      ArgusFree(data->argus);
      data->argus = RaCopyArgusRecord(tmp);

      data->status |= RA_MODIFIED;
      data->argus->ahdr.status |= ARGUS_MERGED;
      data->farhdrstatus = ArgusIndexRecord (data->argus, data->farhdrs);


      if (Rflag) {
         switch (far2->flow.ip_flow.ip_p) {
            case IPPROTO_UDP:
            case IPPROTO_ICMP:
               break;
         }
      }

#ifdef ARGUSDEBUG
      ArgusDebug (5, "RaMergeArgusRecord: merged\n");
#endif
      return;

   } else {
      if (RaThisFlowRev)
         ArgusReverseRecord(a1);

      data->argus = RaCopyArgusRecord(a1);
      data->farhdrstatus = ArgusIndexRecord (data->argus, data->farhdrs);

#ifdef ARGUSDEBUG
      ArgusDebug (5, "RaMergeArgusRecord: swapped");
#endif
      return;
   }

doneRaMergeArgus:
   RaSendArgusRecord(store);

   ArgusFree(data->argus);

   if (RaThisFlowRev)
      ArgusReverseRecord(a1);

   data->argus = RaCopyArgusRecord(a1);
   bcopy ((char *)&RaArgusFlow->ip_flow, (char *)&data->flow.ip_flow, sizeof(RaArgusFlow->ip_flow));

   data->status &= ~((ARGUS_FAR_SRCADDR_MODIFIED | ARGUS_FAR_DSTADDR_MODIFIED) |
                     (ARGUS_FAR_DSTPORT_MODIFIED | ARGUS_FAR_DSTPORT_MODIFIED) |
                     (ARGUS_FAR_PROTO_MODIFIED    | ARGUS_FAR_TPVAL_MODIFIED));
   data->status |= RA_MODIFIED;
   data->farhdrstatus = ArgusIndexRecord (data->argus, data->farhdrs);

#ifdef ARGUSDEBUG
   ArgusDebug (5, "RaMergeArgusRecord: dumped and swapped");
#endif
}


void
RaUpdateArgusStore(struct ArgusRecord *argus, struct ArgusRecordStore *store)
{
   store->status |= RA_MODIFIED;
   store->qhdr.lasttime = ArgusGlobalTime;

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaUpdateArgusStore: \n");
#endif
}


int
RaTimeoutArgusStore (struct ArgusRecordStore *store)
{
   struct ArgusRecord *argus;
   int i, retn = 0;

   if (store != NULL) {
      if (store->status & RA_MODIFIED)
         retn = RaSendArgusRecord(store);

      if (store->rahtblhdr)
         RaRemoveHashEntry (store->rahtblhdr->htable, store->rahtblhdr);

      for (i = 0; i < RaHistoTimeSeries + 1; i++) {
         struct ArgusRecordData *data = NULL;
         
         if ((data = store->data[i]) != NULL) {
            if ((argus = data->argus) != NULL) {
               RaRemoveArgusRecord(argus);
               ArgusFree (argus);
               data->argus = NULL;
            }
            ArgusFree(data);
            store->data[i] = NULL;
         }
      }

      if (store->queue != NULL)
         RaDeleteQueue(store->queue);

      ArgusFree (store->data);
      ArgusFree (store);
      RaAllocArgusRecordStore--;
   }

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaTimeoutArgusStore: 0x%x returning %d\n", store, retn);
#endif

   return (retn);
}


int
RaCheckTimeout (struct ArgusRecordStore *store, struct ArgusRecord *argus)
{
   int retn = 0, lapseTime;
   struct ArgusRecordData *data = NULL;
   struct ArgusRecord *sargus = NULL;

   struct timeval *tvp = &ArgusGlobalTime;

   int lapseTimeSecs, lapseTimeuSecs;
   int starTimeSecs, starTimeuSecs;
   int lastTimeSecs, lastTimeuSecs;

   if (store->ArgusTimeout > 0) {
      lapseTime = store->qhdr.logtime.tv_sec + store->ArgusTimeout;

      if ((tvp->tv_sec > lapseTime) || ((tvp->tv_sec == lapseTime) &&
              (tvp->tv_usec > store->qhdr.logtime.tv_usec))) {
         RaSendArgusRecord(store);

      } else {
         if (argus != NULL) {
            if ((data = store->data[0]) != NULL) {
               sargus = data->argus;

               starTimeSecs  = sargus->argus_far.time.start.tv_sec;
               starTimeuSecs = sargus->argus_far.time.start.tv_usec;
         
               if (argus->argus_far.time.last.tv_sec > sargus->argus_far.time.last.tv_sec) {
                  lastTimeSecs  = argus->argus_far.time.last.tv_sec;
                  lastTimeuSecs = argus->argus_far.time.last.tv_usec;
               } else {
                  lastTimeSecs  = sargus->argus_far.time.last.tv_sec;
                  if (argus->argus_far.time.last.tv_sec == sargus->argus_far.time.last.tv_sec) {
                     if (argus->argus_far.time.last.tv_usec > sargus->argus_far.time.last.tv_usec) {
                        lastTimeuSecs  = argus->argus_far.time.last.tv_usec;
                     } else
                        lastTimeuSecs  = sargus->argus_far.time.last.tv_usec;
                  } else
                     lastTimeuSecs  = sargus->argus_far.time.last.tv_usec;
               }

               lapseTimeSecs  = (lastTimeSecs  - starTimeSecs);
               lapseTimeuSecs = (lastTimeuSecs - starTimeuSecs);

               if (lapseTimeuSecs < 0) {
                  lapseTimeSecs--;
                  lapseTimeuSecs += 1000000;
               }

               if (lapseTimeSecs >= store->ArgusTimeout)
                  RaSendArgusRecord(store);
            }
         }
      }
   }

   if (store->ArgusIdleTimeout > 0) {
      if ((data = store->data[0]) != NULL) {
         sargus = data->argus;
         lastTimeSecs  = sargus->argus_far.time.last.tv_sec;
         lastTimeuSecs = sargus->argus_far.time.last.tv_usec;

         lapseTimeSecs  = (tvp->tv_sec  - lastTimeSecs);
         lapseTimeuSecs = (tvp->tv_usec - lastTimeuSecs);

         if (lapseTimeSecs >= store->ArgusIdleTimeout)
            retn++;
      }
   }

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaCheckTimeout: returning %d \n", retn);
#endif

   return (retn);
}

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


int
RaParseType (char *str)
{
   return(argus_nametoeproto(str));
}


struct RaCIDRAddr *
RaParseCIDRAddr (char *str)
{
   int error = 0, vlen, notIPAddr = 0, i;
   struct RaCIDRAddr *retn = NULL;
   char *ptr = NULL, chr = '\0';
   unsigned int **alist;
   int maskIndex;

   RaCIDRPtr->type = RA_CIDR_TYPE;
   RaCIDRPtr->addr = 0;
   RaCIDRPtr->mask = 0xFFFFFFFF;

   if ((ptr = strchr(str, '*')) == NULL) {
      if ((strncasecmp (str, "class", 5)) != 0) {
         if ((ptr = strchr (str, ':')) != NULL) {
            chr = *ptr;
            *ptr++ = 0;
         } else {
            if ((ptr = strchr (str, '/')) != NULL) {
               chr = *ptr;
               *ptr++ = 0;
            }
         }
   
         for (i = 0; i < strlen(str); i++)
           if (!(isdigit((int)str[i]) || (str[i] == '.')))
              notIPAddr++;
   
         if (!(notIPAddr)) {
            vlen = __argus_atoin(str, &RaCIDRPtr->addr);
   
            /* Promote short net number */
   
            while (RaCIDRPtr->addr && (RaCIDRPtr->addr & 0xff000000) == 0) {
               RaCIDRPtr->addr <<= 8;
               RaCIDRPtr->mask <<= 8;
            }
   
         } else {
            if ((alist = argus_nametoaddr(str)) != NULL)  {
               RaCIDRPtr->addr = **alist;
            } else
               error++;
         }
      
         if (ptr && (((chr == ':') || (chr == '/')) && isdigit((int) *ptr))) {
            maskIndex = atoi(ptr);
      
            if ((maskIndex) > 0) {
               RaCIDRPtr->mask = 0xFFFFFFFF;
               RaCIDRPtr->mask <<= (32 - maskIndex);
            }
   
            if (RaCIDRPtr->mask)
               RaCIDRPtr->addr &= RaCIDRPtr->mask;
         }
      } else {
         RaCIDRPtr->type = RA_NETCLASS_TYPE;
      }
   }

   if (!error)
      retn = RaCIDRPtr;

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaParseCIDRAddr: returning 0x%x \n", retn);
#endif
   
   return (retn);
}


int RaParseProtoField = 1;
struct RaPolicyStruct *
RaParsePolicyEntry (char *str)
{
   struct RaPolicyStruct *retn = NULL;
   struct RaCIDRAddr *addr = NULL;
   int i, type, policyelements, error = 0;
   unsigned short prototype;
   char buf[MAXSTRLEN], *ptr = NULL, *tmp = buf;

   if (strncmp (str, RA_FLOWTAGSTR, strlen(RA_FLOWTAGSTR)) == 0) {
      type = RA_FLOWLIST;
      policyelements = RA_FLOWPOLICYFIELDNUM;

   } else {
      if ((strncmp (str, RA_MODELTAGSTR, strlen(RA_MODELTAGSTR))) == 0) {
         type = RA_MODELIST;
         policyelements = RA_MODELPOLICYFIELDNUM;

      } else 
         return (NULL);
   }

   if ((retn = (struct RaPolicyStruct *) ArgusCalloc (1, sizeof (*retn))) == NULL)
      return retn;

   bzero (buf, MAXSTRLEN);
   bcopy ((char *) str, buf, strlen(str));

   for (i = 0; ((i < policyelements) && !(error)); i++) {
      if ((ptr = strtok (tmp, " \t")) && (*ptr != '\n')) {
         switch (i) {
            case RA_LABELSTRING: 
               switch (type) {
                  case RA_FLOWLIST:
                     if (strcmp (ptr,  RA_FLOWTAGSTR))
                        error++;
                     else 
                        retn->RaEntryType = RA_FLOWLIST;
                     break;

                  case RA_MODELIST:
                     if (strcmp (ptr, RA_MODELTAGSTR))
                        error++;
                     else 
                        retn->RaEntryType = RA_MODELIST;
                     break;
               }
               break;
   
            case RA_POLICYID: 
               if (!(retn->RaPolicyId = atoi (ptr))) error++;
               break;

            case RA_POLICYTYPE:
               if ((prototype = RaParseType (ptr)) != 0)
                  retn->type = prototype & 0xFFFF;
               else
                  error++;
               break;
   
            case RA_POLICYSRCADDR:
               if ((addr = RaParseCIDRAddr (ptr)) != NULL)
                  retn->src = *addr;
               else
                  error++;
               break;
   
            case RA_POLICYDSTADDR:
               if ((addr = RaParseCIDRAddr (ptr)) != NULL)
                  retn->dst = *addr;
               else
                  error++;
               break;
   
            case RA_POLICYPROTO: 
               if (RaParseProtoField) {
                  switch (type) {
                     case RA_FLOWLIST:
                        if (isdigit((int)*ptr)) {
                           retn->proto = atoi(ptr);
                        } else {
                           if (*ptr == '*') {
                              retn->proto = 0xFF;
                           } else {
                              struct protoent *proto;
                              if ((proto = getprotobyname(ptr)) != NULL)
                                 retn->proto = proto->p_proto;
                              else {
                                 error++;
                              }
                           }
                        }
                        break;

                     case RA_MODELIST:
                        if ((strstr(ptr, "yes")) != NULL)
                           retn->proto = 0xFF;
                        else
                           if ((strstr(ptr, "no")) != NULL)
                              retn->proto = 0x00;
                           else
                              error++;
                        break;
                  }
               }
               break;

            case RA_POLICYSRCPORT:
               switch (type) {
                  case RA_FLOWLIST:
                     if (isdigit((int)*ptr)) {
                        retn->sport = (arg_uint16) atoi (ptr);
                     } else {
                        if (*ptr == '*')
                           retn->sport = 0xFFFF;
                        else {
                           int proto = retn->proto, port = 0;
                           if (retn->proto == IPPROTO_ICMP) {
                              if (!(strcmp (ptr, "echo"))) retn->sport = ICMP_ECHO; else
                              if (!(strcmp (ptr, "unreach"))) retn->sport = ICMP_UNREACH; else
                              if (!(strcmp (ptr, "squench"))) retn->sport = ICMP_SOURCEQUENCH; else
                              if (!(strcmp (ptr, "redirect"))) retn->sport = ICMP_REDIRECT; else
                              if (!(strcmp (ptr, "timexed"))) retn->sport = ICMP_TIMXCEED; else
                              if (!(strcmp (ptr, "param"))) retn->sport = ICMP_PARAMPROB; else
                              if (!(strcmp (ptr, "timestmp"))) retn->sport = ICMP_TSTAMP; else
                              if (!(strcmp (ptr, "info"))) retn->sport = ICMP_IREQ; else
                              if (!(strcmp (ptr, "advert"))) retn->sport = ICMP_ROUTERADVERT; else
                              if (!(strcmp (ptr, "solicit"))) retn->sport = ICMP_ROUTERSOLICIT; else
                                 error++;
                           } else {
                              if (argus_nametoport(ptr, &port, &proto)) {
                                 retn->sport = port;
                              } else
                                 error++;
                           }
                        }
                     }
                     break;

                  case RA_MODELIST:
                     if ((strstr(ptr, "yes")) != NULL)
                        retn->sport = 0xFFFF;
                     else
                        if ((strstr(ptr, "no")) != NULL)
                           retn->sport = 0x0000;
                        else
                           error++;
                     break;
               }
               break;
   
            case RA_POLICYDSTPORT:
               switch (type) {
                  case RA_FLOWLIST:
                     if (isdigit((int)*ptr)) {
                        retn->dport = (arg_uint16) atoi (ptr);
                     } else {
                        if (*ptr == '*')
                           retn->dport = 0xFFFF;
                        else {
                           if (retn->proto == IPPROTO_ICMP) {
                           } else {
                              int proto = retn->proto, port;
                              if (argus_nametoport(ptr, &port, &proto)) {
                                 retn->dport = port;
                              } else
                                 error++;
                           }
                        }
                     }
                     break;

                  case RA_MODELIST:
                     if ((strstr(ptr, "yes")) != NULL)
                        retn->dport = 0xFFFF;
                     else
                        if ((strstr(ptr, "no")) != NULL)
                           retn->dport = 0x0000;
                        else
                           error++;
                     break;
               }
               break;
      
            case RA_POLICYMODELST:
               if (type == RA_FLOWLIST) {
                  if (isdigit((int)*ptr))
                     retn->RaModelId = (arg_uint32) atoi (ptr);
                  else
                     error++;
               }
               break;

            case RA_POLICYTIMEOUT:
               if (type == RA_FLOWLIST) {
                  if (isdigit((int)*ptr))
                     retn->ArgusTimeout = (arg_uint32) atoi (ptr);
                  else
                     error++;
               }

            case RA_POLICYIDLETIMEOUT:
               if (type == RA_FLOWLIST) {
                  if (isdigit((int)*ptr))
                     retn->ArgusIdleTimeout = (arg_uint32) atoi (ptr);
               }
               break;
         }
         tmp = NULL;

      } else
         break;
   }

   if (error) {
      ArgusFree(retn);
      retn = NULL;
   }

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaParsePolicyEntry: returning 0x%x \n", retn);
#endif

   return (retn);
}


int
RaCreatePolicyEntry (struct RaFlowModelStruct *fmodel, int index, char *str)
{
   int retn = 0;
   struct RaPolicyStruct *rap;

   if (strncmp (str, RA_MODELNAMETAGSTR, strlen(RA_MODELNAMETAGSTR)) == 0) {
      fmodel->desc = strdup(&str[strlen(RA_MODELNAMETAGSTR)]);
      return retn;
   }

   if (strncmp (str, RA_PRESERVETAGSTR, strlen(RA_PRESERVETAGSTR)) == 0) {
      if (!(strncmp (&str[strlen(RA_PRESERVETAGSTR)], "yes", 3))) {
         fmodel->preserve = 1;
      } else
         RaPrintActualValues = 0;
      return retn;
   }

   if (strncmp (str, RA_REPORTTAGSTR, strlen(RA_REPORTTAGSTR)) == 0) {
      if (!(strncmp (&str[strlen(RA_REPORTTAGSTR)], "yes", 3)))
         fmodel->report = 1;
      return retn;
   }

   if (strncmp (str, RA_AUTOCORRECTSTR, strlen(RA_AUTOCORRECTSTR)) == 0) {
      if (!(strncmp (&str[strlen(RA_AUTOCORRECTSTR)], "yes", 3))) {
         fmodel->autocorrect = 1;
      } else {
         fmodel->autocorrect = 0;
         RaAutoCorrection = 0;
      }
      return retn;
   }

   if ((rap = RaParsePolicyEntry (str)) != NULL) {
      rap->str = str;

      switch (rap->RaEntryType) {
         case RA_FLOWLIST:
            fmodel->policy[fmodel->pindex++] = rap;
            break;

         case RA_MODELIST:
            fmodel->model[fmodel->mindex++] = rap;
            break;
      }
   } else
      ArgusLog (LOG_ERR, "RaCreatePolicyEntry: format error line %d %s", index, str);

#ifdef ARGUSDEBUG
   ArgusDebug (6, "RaCreatePolicyEntry: returning %d\n", retn);
#endif

   return (retn);
}


#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#define ARGUSMAXMODELFILE	2048
#define RAMAXPOLICYSIZE		2048
char *modelbuf[ARGUSMAXMODELFILE];


struct RaFlowModelStruct *
RaReadFlowModelFile (char **model)
{
   struct RaFlowModelStruct *retn = NULL;
   int RaPolicyLines = 0, RaPolicyLineNumber = 0;
   char strbuf[MAXSTRLEN], *str = strbuf;
   int i = 0, x = 0;
   FILE *fd;

   if (model == NULL) {
      bzero ((char *) modelbuf, sizeof(modelbuf));
      if ((fd = fopen (ArgusFlowModelFile, "r")) != NULL) {
         while ((str = fgets (str, MAXSTRLEN, fd)) != NULL)
            modelbuf[i++] = strdup(str);

         model = modelbuf;

      } else
         ArgusLog (LOG_ERR, "%s: %s", ArgusFlowModelFile, strerror(errno));
   }

   if ((retn = (struct RaFlowModelStruct *) ArgusCalloc (1, sizeof (*retn))) == NULL)
      ArgusLog (LOG_ERR, "RaReadFlowModelFile: ArgusCalloc %s", strerror(errno));

   if ((retn->policy = (struct RaPolicyStruct **) ArgusCalloc (RAMAXPOLICYSIZE, sizeof (void *))) == NULL)
      ArgusLog (LOG_ERR, "RaReadFlowModelFile: ArgusCalloc %s", strerror(errno));

   if ((retn->model = (struct RaPolicyStruct **) ArgusCalloc (RAMAXPOLICYSIZE, sizeof (void *))) == NULL)
      ArgusLog (LOG_ERR, "RaReadFlowModelFile: ArgusCalloc %s", strerror(errno));

   while ((str = *model++) != NULL) {
      RaPolicyLineNumber++;

      while (isspace((int)*str))
         str++;
      if (strlen(str)) {
         switch (*str) {
            case '#':
            case '\n':
            case '!':
               break;

            default:
               if ((str = strdup(str)) != NULL) {
                  if (RaCreatePolicyEntry(retn, RaPolicyLines, str) < 0) {
                     ArgusLog (LOG_ERR, "RaReadFlowModelFile: parse error line %d %s\n", 
                                RaPolicyLineNumber, RaPolicyError[RaParseError]);
                  } else
                     RaPolicyLines++;
                 
               } else
                  ArgusLog (LOG_ERR, "RaReadFlowModelFile: error %s\n", strerror(errno));
         }
      }
   }

   for (i = 0; i < RaThisFlowNum; i++) {
      if (retn->policy[i]) {
         int found = 0;
         for (x = 0; x < RaThisModelNum; x++) {
            if (retn->model[x])
               if (retn->policy[i]->RaModelId == retn->model[x]->RaPolicyId) {
                  retn->policy[i]->RaModelId = x;
                  found++;
                  break;
               }
         }
         if (!found)
            ArgusLog (LOG_ERR, "%s: RaReadFlowModelFile: Model Index %d not found\n",
                             ArgusProgramName, retn->policy[i]->RaModelId);
      }
   }

   if (modelbuf[0] != NULL) {
      do {
         free(modelbuf[i]);
         modelbuf[i++] = NULL;
      } while (modelbuf[i] != NULL);
   }

#ifdef ARGUSDEBUG
   ArgusDebug (3, "RaReadFlowModelFile: returning 0x%x\n", retn);
#endif

   return (retn);
}

struct RaHashTableHeader *
RaFindHashEntry (struct RaHashTableStruct *thisHashTable, struct RaHashStruct *hstruct)
{
   struct RaHashTableHeader *retn = NULL;
   struct RaHashTableHeader *head = NULL, *target;

   if (thisHashTable->size > 0) {
      if ((target = thisHashTable->array[hstruct->hash % thisHashTable->size]) != NULL) {
         head = target;
         do {
            if (!(bcmp ((char *) hstruct->buf, (char *) &target->hstruct.buf, sizeof(hstruct->buf)))) {
                  retn = target;
                  break;
               } else
                  target = target->nxt;
         } while (target != head);
      }
   }

#ifdef ARGUSDEBUG
   ArgusDebug (7, "RaFindHashEntry: returning 0x%x\n", retn);
#endif

   return (retn);
}



struct RaHashTableHeader *
RaAddHashEntry (struct RaHashTableStruct *thisHashTable, struct ArgusRecordStore *store)
{
   struct RaHashTableHeader *retn = NULL;
   struct RaHashTableHeader *start = NULL;
   struct ArgusFlow *RaThisFlow = NULL;
   struct ArgusRecord *argus = NULL;
   struct RaHashStruct ArgusHash;
   unsigned short ip_id = 0, *ptr = NULL;
   int i, len;


   if (store && store->data && store->data[RaThisActiveIndex] && (argus = store->data[RaThisActiveIndex]->argus)) {
      if ((retn = (struct RaHashTableHeader *) ArgusCalloc (1, sizeof (struct RaHashTableHeader))) != NULL) {
         RaAllocHashTableHeaders++;
         RaThisFlow = &argus->argus_far.flow;

         bzero ((char *)&ArgusHash, sizeof(ArgusHash));
         ArgusHash.len = sizeof(ArgusHash.buf);

         if ((argus->ahdr.status & 0xFFFF) == ETHERTYPE_IP) {
            switch (argus->argus_far.flow.ip_flow.ip_p) {

               case IPPROTO_ICMP:
                  ip_id = RaThisFlow->icmp_flow.ip_id;
                  RaThisFlow->icmp_flow.ip_id = 0;
                  break;

               default:
                  ip_id = RaThisFlow->ip_flow.ip_id;
                  RaThisFlow->ip_flow.ip_id = 0;
                  break;
            }
         }

         ArgusHash.buf[0] = argus->ahdr.argusid;
         bcopy (RaThisFlow, (char *)&ArgusHash.buf[1], sizeof(*RaThisFlow));
         ptr = (unsigned short *)&ArgusHash.buf[0];

         for (i = 0, len = ArgusHash.len / sizeof(unsigned short); i < len; i++)
            ArgusHash.hash += *ptr++;

         bcopy((char *)&ArgusHash, (char *)&retn->hstruct, sizeof (ArgusHash));
         retn->storeobj = store;
         
         if ((start = thisHashTable->array[ArgusHash.hash % thisHashTable->size]) != NULL) {
            retn->nxt = start;
            retn->prv = start->prv;
            retn->prv->nxt = retn;
            retn->nxt->prv = retn;
         } else
            retn->prv = retn->nxt = retn;

         thisHashTable->array[ArgusHash.hash % thisHashTable->size] = retn;
         thisHashTable->count++;
         retn->htable = thisHashTable;

         if ((argus->ahdr.status & 0xFFFF) == ETHERTYPE_IP) {
            switch (argus->argus_far.flow.ip_flow.ip_p) {

               case IPPROTO_ICMP:
                  RaThisFlow->icmp_flow.ip_id = ip_id;
                  break;

               default:
                  RaThisFlow->ip_flow.ip_id = ip_id;
                  break;
            }
         }
      }
   }

#ifdef ARGUSDEBUG
   ArgusDebug (7, "RaAddHashEntry (0x%x) returning 0x%x\n", store, retn);
#endif

   return (retn);
}

void
RaRemoveHashEntry (struct RaHashTableStruct *thisHashTable, struct RaHashTableHeader *rahtblhdr)
{
   unsigned short hash = rahtblhdr->hstruct.hash;

   if (rahtblhdr->htable == thisHashTable) {
      if ((rahtblhdr->prv)  && (rahtblhdr->nxt)) {
         rahtblhdr->prv->nxt = rahtblhdr->nxt;
         rahtblhdr->nxt->prv = rahtblhdr->prv;

         if ((rahtblhdr->nxt == rahtblhdr) || (rahtblhdr->prv == rahtblhdr)) {
            rahtblhdr->nxt = NULL;
            rahtblhdr->prv = NULL;
         }
      }

      rahtblhdr->htable = NULL;

      if (thisHashTable) {
         if (rahtblhdr == thisHashTable->array[hash % thisHashTable->size]) {
            if (rahtblhdr == rahtblhdr->nxt)
               thisHashTable->array[hash % thisHashTable->size] = NULL;
            else
               thisHashTable->array[hash % thisHashTable->size] = rahtblhdr->nxt;
         }

         thisHashTable->count--;

      }
   } else
      ArgusLog (LOG_ERR, "RaRemoveHashEntry (0x%x, 0x%x) obj not in table\n", thisHashTable, rahtblhdr);

   ArgusFree (rahtblhdr);
   RaAllocHashTableHeaders--;
 
#ifdef ARGUSDEBUG
   ArgusDebug (7, "RaRemoveHashObject (0x%x) returning\n", rahtblhdr);
#endif
}

void
RaCleanHashTable (struct RaHashTableStruct *thisHashTable)
{
   struct RaHashTableHeader *rahtblhdr = NULL;
   int i;

   for (i = 0; i < thisHashTable->size; i++)
      while ((rahtblhdr = thisHashTable->array[i]) != NULL)
         RaRemoveHashEntry(rahtblhdr->htable, rahtblhdr);

   ArgusFree (thisHashTable->array);
   thisHashTable->array = NULL;

#ifdef ARGUSDEBUG
   ArgusDebug (7, "RaCleanHashTable (0x%x) returning\n", thisHashTable);
#endif
}

void
RaDeleteHashTable (struct RaHashTableStruct *thisHashTable)
{
   struct RaHashTableHeader *rahtblhdr = NULL;
   int i;

   for (i = 0; i < thisHashTable->size; i++)
      while ((rahtblhdr = thisHashTable->array[i]) != NULL)
         RaRemoveHashEntry(rahtblhdr->htable, rahtblhdr);

   if (thisHashTable->array != NULL)
      ArgusFree (thisHashTable->array);

   ArgusFree (thisHashTable);

#ifdef ARGUSDEBUG
   ArgusDebug (7, "RaDeleteHashTable (0x%x) returning\n", thisHashTable);
#endif
}



struct RaQueueStruct *
RaNewQueue ()
{
   struct RaQueueStruct *retn =  NULL;

   if ((retn = (struct RaQueueStruct *) ArgusCalloc (1, sizeof (struct RaQueueStruct))) != NULL) {
      retn->count = 0;
      retn->start = NULL;
   }

#ifdef ARGUSDEBUG
   ArgusDebug (4, "RaNewQueue () returning 0x%x\n", retn);
#endif

   return (retn);
}

void
RaDeleteQueue (struct RaQueueStruct *queue)
{
   struct ArgusQueueHeader *obj = NULL;

   if (queue != NULL) {
      while ((obj = RaPopQueue(queue)))
         ArgusFree(obj);

      ArgusFree(queue);
   }

#ifdef ARGUSDEBUG
   ArgusDebug (4, "RaDeleteQueue (0x%x) returning\n", queue);
#endif
}


int
RaGetQueueCount(struct RaQueueStruct *queue)
{

#ifdef ARGUSDEBUG
   ArgusDebug (7, "RaGetQueueCount (0x%x) returning %d\n", queue, queue->count);
#endif

   return (queue->count);
}

void
RaPushQueue(struct RaQueueStruct *queue, struct ArgusQueueHeader *obj)
{
   int retn = 0;

   if ((retn = RaAddToQueue (queue, obj)) > 0)
      queue->start = queue->start->prv;

#ifdef ARGUSDEBUG
   ArgusDebug (7, "RaPushQueue (0x%x, 0x%x) returning\n", queue, obj);
#endif
}


int
RaAddToQueue(struct RaQueueStruct *queue, struct ArgusQueueHeader *obj)
{
   int retn = 0;

   if (queue && obj) {
      if (obj->queue == NULL) {
         if (queue->start != NULL) {
            obj->prv = queue->start->prv;
            queue->start->prv = obj;
            obj->nxt = queue->start;
            obj->prv->nxt = obj;
         } else {
            queue->start = obj;
            obj->nxt = obj;
            obj->prv = obj;
         }
         queue->count++;
         obj->queue = queue;
         retn = 1;

      } else
         ArgusLog (LOG_ERR, "RaAddToQueue (0x%x, 0x%x) obj in queue 0x%x\n", queue, obj, obj->queue);
   } else
      ArgusLog (LOG_ERR, "RaAddToQueue (0x%x, 0x%x) parameter error\n", queue, obj);

#ifdef ARGUSDEBUG
   ArgusDebug (7, "RaAddToQueue (0x%x, 0x%x) returning %d\n", queue, obj, retn);
#endif

   return (retn);
}


struct ArgusQueueHeader *
RaPopQueue (struct RaQueueStruct *queue)
{
   struct ArgusQueueHeader *retn = NULL;
   struct ArgusQueueHeader *obj = NULL;

   if (queue && queue->count) {
      if ((obj = (struct ArgusQueueHeader *) queue->start) != NULL)
         retn = RaRemoveFromQueue(queue, obj);
      else
         ArgusLog (LOG_ERR, "RaPopQueue(0x%x) internal queue error count %d start NULL\n", queue, queue->count);
   }

#ifdef ARGUSDEBUG
   ArgusDebug (8, "RaPopQueue (0x%x) returning 0x%x\n", queue, retn);
#endif
   
   return(retn);
}


struct ArgusQueueHeader *
RaRemoveFromQueue(struct RaQueueStruct *queue, struct ArgusQueueHeader *obj)
{
   struct ArgusQueueHeader *retn = NULL;

   if ((queue != NULL) && (obj != NULL)) {
      if (queue->count && (obj->queue == queue)) {
         queue->count--;

         if (queue->count) {
            if (queue->start == obj)
               queue->start = obj->nxt;

            obj->prv->nxt = obj->nxt;
            obj->nxt->prv = obj->prv;

         } else
            queue->start = NULL;

         obj->prv = NULL;
         obj->nxt = NULL;
         obj->queue = NULL;
         retn = obj;

      } else
         ArgusLog (LOG_ERR, "RaRemoveFromQueue(0x%x, 0x%x) obj not in queue\n", queue, obj);
   } else
      ArgusLog (LOG_ERR, "RaRemoveFromQueue(0x%x, 0x%x) parameter error\n", queue, obj);

#ifdef ARGUSDEBUG
   ArgusDebug (9, "RaRemoveFromQueue (0x%x, 0x%x) returning 0x%x\n", queue, obj, obj);
#endif

   return (retn);
}


struct ArgusListStruct *
ArgusNewList ()
{
   struct ArgusListStruct *retn = NULL;
 
   if ((retn = (struct ArgusListStruct *) ArgusCalloc (1, sizeof (struct ArgusListStruct))) != NULL) {
      retn->start = NULL;
      retn->count = 0;
   }

#ifdef ARGUSDEBUG
   ArgusDebug (4, "ArgusNewList () returning 0x%x\n", retn);
#endif

   return (retn);
}

void
ArgusDeleteList (struct ArgusListStruct *list)
{
   if (list) {
      while (list->start)
         ArgusPopFrontList(list);

      ArgusFree (list);
   }

#ifdef ARGUSDEBUG
   ArgusDebug (4, "ArgusDeleteList (0x%x) returning\n", list);
#endif
}

int
ArgusListEmpty (struct ArgusListStruct *list)
{
   return (list->start == NULL);
}

int
ArgusGetListCount(struct ArgusListStruct *list)
{
   return (list->count);
}

void
ArgusPushFrontList(struct ArgusListStruct *list, void *obj)
{
   struct ArgusListObjectStruct *lobj = NULL;

   if (list && obj) {
      if ((lobj = (struct ArgusListObjectStruct *) ArgusCalloc (1, sizeof(*lobj))) != NULL) {
         lobj->obj = obj;
   
         if (list->start) {
            lobj->nxt = list->start;
            lobj->prv = list->start->prv;
            lobj->nxt->prv = lobj;
            lobj->prv->nxt = lobj;
         } else {
            lobj->prv = lobj;
            lobj->nxt = lobj;
         }
   
         list->start = lobj;
         list->count++;
      }
   }

#ifdef ARGUSDEBUG
   ArgusDebug (7, "ArgusPushFrontList (0x%x, 0x%x) returning 0x%x\n", list, obj);
#endif
}

void
ArgusPushBackList(struct ArgusListStruct *list, void *obj)
{
   ArgusPushFrontList(list, obj);
   list->start = list->start->nxt;

#ifdef ARGUSDEBUG
   ArgusDebug (7, "ArgusPushBackList (0x%x, 0x%x) returning 0x%x\n", list, obj);
#endif
}


void *
ArgusFrontList(struct ArgusListStruct *list)
{
   void *retn = NULL;

   if (list->start)
      retn = list->start->obj;

#ifdef ARGUSDEBUG
   ArgusDebug (7, "ArgusFrontList (0x%x) returning 0x%x\n", list, retn);
#endif

   return (retn);
}

void *
ArgusBackList(struct ArgusListStruct *list)
{
   void *retn = NULL;

   if (list->start)
      retn = list->start->prv->obj;

#ifdef ARGUSDEBUG
   ArgusDebug (6, "ArgusBackList (0x%x) returning 0x%x\n", list, retn);
#endif

   return (retn);
}



void
ArgusPopBackList(struct ArgusListStruct *list)
{
   struct ArgusListObjectStruct *lobj = NULL;

   if ((lobj = list->start)) {
      list->start = list->start->prv;
      ArgusPopFrontList(list);
   }

#ifdef ARGUSDEBUG
   ArgusDebug (7, "ArgusPopBackList (0x%x) returning\n", list);
#endif
}

void
ArgusPopFrontList(struct ArgusListStruct *list)
{
   struct ArgusListObjectStruct *lobj = NULL;

   if ((lobj = list->start)) {
      if (--list->count > 0) {
         if (lobj->prv)
            lobj->prv->nxt = lobj->nxt;
 
         if (lobj->nxt)
            lobj->nxt->prv = lobj->prv;
 
        list->start = lobj->nxt;
 
      } else
         list->start = NULL;
 
      ArgusFree(lobj);
   }

#ifdef ARGUSDEBUG
   ArgusDebug (7, "ArgusPopFrontList (0x%x) returning\n", list);
#endif
}

char RaPrecisionPad[32], RaTimePad[32], RaDateBuf[128];

char *
RaGenerateLabel(struct ArgusRecord *argus) 
{
   int i, x;

   bzero (RaDateBuf, sizeof (RaDateBuf));
   bzero (RaTimePad, sizeof (RaTimePad));
   bzero (RaPrecisionPad, sizeof (RaPrecisionPad));
   bzero (RaLabelStr, sizeof(RaLabelStr));

   ArgusPrintStartDate (RaDateBuf, argus);
   if (strlen(RaDateBuf) > 10) {
      for (x = 0; x < (strlen(RaDateBuf) - 10)/2; x++)
         strcat (RaTimePad, " ");
   }
    
   if (pflag)
      for (x = 0; x < pflag/2; x++)
         strcat (RaPrecisionPad, " ");

   RaLabel = RaLabelStr;

   for (i = 0; i < MAX_PRINT_ALG_TYPES; i++) {
      if (RaPrintAlgorithms[i] != NULL) {
         for (x = 0; x < MAX_PRINT_ALG_TYPES; x++) {
            if ((void *) RaPrintAlgorithms[i] == (void *) RaPrintAlgorithmTable[x]) {
               RaPrintLabelAlgorithmTable[x](&RaLabel[strlen(RaLabel)]);
               break;
            }
         }
      } else
         break;
   }

   if ((RaFieldDelimiter != ' ') && (RaFieldDelimiter != '\0')) {
      char tmpbuf[1024], *ptr = tmpbuf, *str = RaLabel, lastchr = ' ';

      bzero (tmpbuf, sizeof(tmpbuf));
      lastchr = RaFieldDelimiter;
      while (*str) {
         if (*str == ' ') {
            if (lastchr != RaFieldDelimiter)
               *ptr++ = RaFieldDelimiter;
            while (isspace((int)*str)) str++;
         }
         lastchr = *str;

         *ptr++ = *str++;
      }
      bzero (RaLabel, MAXSTRLEN);
      if (tmpbuf[strlen(tmpbuf) - 1] == RaFieldDelimiter)
         tmpbuf[strlen(tmpbuf) - 1] = '\0';

      bcopy (tmpbuf, RaLabel, strlen(tmpbuf));
   }

   return (RaLabel);
}


void
ArgusPrintDateLabel (char *buf)
{
   if (lflag)
      ArgusPrintLastDateLabel (buf);
   else {
      ArgusPrintStartDateLabel (buf);
      ArgusPrintLastDateLabel (buf);
   }
}


void
ArgusPrintStartDateLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], RaTimePad);
   sprintf (&buf[strlen(buf)], "StartTime ");

   if (pflag) {
      if (!uflag)
         if ((RaFieldDelimiter != ' ') && (RaFieldDelimiter != '\0'))
            sprintf (&buf[strlen(buf)], "Fraction ");
   }

   sprintf (&buf[strlen(buf)], RaTimePad);
   if (strlen(RaDateBuf) & 0x01)
      sprintf (&buf[strlen(buf)], " ");
}

void
ArgusPrintLastDateLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], RaTimePad);
   sprintf (&buf[strlen(buf)], " LastTime ");

   if (pflag) {
      if (!uflag)
         if ((RaFieldDelimiter != ' ') && (RaFieldDelimiter != '\0'))
            sprintf (&buf[strlen(buf)], "Fraction ");
   }

   sprintf (&buf[strlen(buf)], RaTimePad);
   if (strlen(RaDateBuf) & 0x01)
      sprintf (&buf[strlen(buf)], " ");
}

void
ArgusPrintSourceIDLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], "     ProbeId    ");
}

void
ArgusPrintIndicatorLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], "  Flgs   ");
}

void
ArgusPrintMACAddressLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], RamOptionLabel);
}

void
ArgusPrintProtoLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], "Type ");
}

void
ArgusPrintSrcNetLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], "%*sSrcNet%*s ", (hfield - 6)/2, " ", (hfield - 6)/2, " ");
   if ((hfield & 0x01)) sprintf (&buf[strlen(buf)], " ");
}


void
ArgusPrintSrcAddrLabel (char *buf)
{
   if (ArgusInput->ArgusManStart.ahdr.type & ARGUS_RMON)
      sprintf (&buf[strlen(buf)], "%*s Addr%*s ",   (hfield - 5)/2, " ", (hfield - 5)/2, " ");
   else
      sprintf (&buf[strlen(buf)], "%*sSrcAddr%*s ", (hfield - 7)/2, " ", (hfield - 7)/2, " ");

   if (!(hfield & 0x01)) sprintf (&buf[strlen(buf)], " ");
}

void
ArgusPrintDstNetLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], " %*sDstNet%*s ", (hfield - 6)/2, " ", (hfield - 6)/2, " ");
   if ((hfield & 0x01)) sprintf (&buf[strlen(buf)], " ");
}

void
ArgusPrintDstAddrLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], "%*sDstAddr%*s ", (hfield - 7)/2, " ", (hfield - 7)/2, " ");
   if (!(hfield & 0x01)) sprintf (&buf[strlen(buf)], " ");
}

void
ArgusPrintSrcPortLabel (char *buf)
{
   if (!(pfield & 0x01))
      sprintf (&buf[strlen(buf)], " ");
   if (pfield > 6) {
      sprintf (&buf[strlen(buf)], "%*sSport%*s ", (pfield - 5)/2, " ", (pfield - 5)/2, " ");
   } else {
      sprintf (&buf[strlen(buf)], "Sport ");
   }
}

void
ArgusPrintDstPortLabel (char *buf)
{
   if (!(pfield & 0x01))
      sprintf (&buf[strlen(buf)], " ");
   if (pfield > 6) {
      sprintf (&buf[strlen(buf)], "%*sDport%*s ", (pfield - 5)/2, " ", (pfield - 5)/2, " ");
   } else {
      sprintf (&buf[strlen(buf)], "Dport ");
   }
}

void
ArgusPrintIpIdLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], " IpId  ");
}

void
ArgusPrintDirLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], "Dir ");
}

void
ArgusPrintCountsLabel (char *buf)
{
   if (ArgusInput->ArgusManStart.ahdr.type & ARGUS_RMON)
      sprintf (&buf[strlen(buf)], "InPkt    OutPkt    ");
   else
      sprintf (&buf[strlen(buf)], "SrcPkt   DstPkt    ");
}

void
ArgusPrintSrcPacketsLabel (char *buf)
{
   if (ArgusInput->ArgusManStart.ahdr.type & ARGUS_RMON)
      sprintf (&buf[strlen(buf)], "InPkt    ");
   else
      sprintf (&buf[strlen(buf)], "SrcPkt   ");
}

void
ArgusPrintDstPacketsLabel (char *buf)
{
   if (ArgusInput->ArgusManStart.ahdr.type & ARGUS_RMON)
      sprintf (&buf[strlen(buf)], "OutPkt    ");
   else
      sprintf (&buf[strlen(buf)], "DstPkt    ");
}

void
ArgusPrintBytesLabel (char *buf)
{
   if (ArgusInput->ArgusManStart.ahdr.type & ARGUS_RMON) {
      if (Aflag)
         sprintf (&buf[strlen(buf)], "InAppBytes   OutAppBytes ");
      else
      if (Rflag)
         sprintf (&buf[strlen(buf)], "Response Information     ");
      else
         sprintf (&buf[strlen(buf)], "InBytes      OutBytes    ");
   } else {
      if (Aflag)
         sprintf (&buf[strlen(buf)], "SAppBytes    DAppBytes   ");
      else
      if (Rflag)
         sprintf (&buf[strlen(buf)], "Response Information     ");
      else
         sprintf (&buf[strlen(buf)], "SrcBytes     DstBytes    ");
   }
}

void
ArgusPrintSrcBytesLabel (char *buf)
{
   if (ArgusInput->ArgusManStart.ahdr.type & ARGUS_RMON) {
      if (Aflag)
         sprintf (&buf[strlen(buf)], "InAppBytes  ");
      else
      if (Rflag)
         sprintf (&buf[strlen(buf)], "InBytes     ");
      else
         sprintf (&buf[strlen(buf)], "InBytes     ");
   } else {
      if (Aflag)
         sprintf (&buf[strlen(buf)], "SAppBytes   ");
      else
      if (Rflag)
         sprintf (&buf[strlen(buf)], "SrcBytes    ");
      else
         sprintf (&buf[strlen(buf)], "SrcBytes    ");
   }
}

void
ArgusPrintDstBytesLabel (char *buf)
{
   if (ArgusInput->ArgusManStart.ahdr.type & ARGUS_RMON) {
      if (Aflag)
         sprintf (&buf[strlen(buf)], "OutAppBytes ");
      else
      if (Rflag)
         sprintf (&buf[strlen(buf)], "OutBytes    ");
      else
         sprintf (&buf[strlen(buf)], "OutBytes    ");
   } else {
      if (Aflag)
         sprintf (&buf[strlen(buf)], "DAppBytes   ");
      else
      if (Rflag)
         sprintf (&buf[strlen(buf)], "SrcBytes    ");
      else
         sprintf (&buf[strlen(buf)], "DstBytes    ");
   }
}

void
ArgusPrintSrcJitterLabel (char *buf)
{
   RaPrintSrcTime = 1;
   RaPrintDstTime = 0;
   ArgusPrintJitterLabel (buf);
}

void
ArgusPrintDstJitterLabel (char *buf)
{
   RaPrintSrcTime = 0;
   RaPrintDstTime = 1;
   ArgusPrintJitterLabel (buf);
}

void
ArgusPrintJitterLabel (char *buf)
{
   if (RaPrintSrcTime) {
      if ((RaPrintActTime) && (RaPrintIdleTime)) {
         sprintf (&buf[strlen(buf)],    "  SrcJitter  ");
      } else {
         if (RaPrintActTime)
            sprintf (&buf[strlen(buf)], "  SrcAJitter ");
         else
         if (RaPrintIdleTime)
            sprintf (&buf[strlen(buf)], "  SrcIJitter ");
         else
            sprintf (&buf[strlen(buf)], "             ");
      }
   }

   if (RaPrintDstTime) {
      if ((RaPrintActTime) && (RaPrintIdleTime))
         sprintf (&buf[strlen(buf)],    "  DstJitter  ");
      else {
         if (RaPrintActTime)
            sprintf (&buf[strlen(buf)], "  DstAJitter ");

         else
      if (RaPrintIdleTime)
            sprintf (&buf[strlen(buf)], "  DstIJitter ");
         else
            sprintf (&buf[strlen(buf)], "             ");
      }
   }
}

void
ArgusPrintStatusLabel (char *buf)
{
   if (zflag || Zflag) {
      sprintf (&buf[strlen(buf)], " Status");
   } else {
      sprintf (&buf[strlen(buf)], "State");
   }
}

void
ArgusPrintUserDataLabel (char *buf)
{
   int len = 0;

   if ((ArgusSrcUserDataLen) || (ArgusDstUserDataLen)) {
      if (ArgusSrcUserDataLen > 0) {
         switch (eflag) {
            case ARGUS_ENCODE_ASCII:
               len = ArgusSrcUserDataLen;
               break;

            case ARGUS_ENCODE_32:
            case ARGUS_ENCODE_64:
               len = ArgusSrcUserDataLen;
               len *= 2;
               break;
         }
         if (ArgusSrcUserDataLen > 10) len++;
         sprintf (&buf[strlen(buf)], " %*ssrcUdata%*s ", (len)/2, " ", (len)/2, " ");
         if (len & 0x01)
            sprintf(&buf[strlen(buf)], " ");

      } else
         sprintf(&buf[strlen(buf)], " ");

      if (ArgusDstUserDataLen > 0) {
         switch (eflag) {
            case ARGUS_ENCODE_ASCII:
               len = ArgusDstUserDataLen;
               break;

            case ARGUS_ENCODE_32:
            case ARGUS_ENCODE_64:
               len = ArgusDstUserDataLen;
               len *= 2;
               break;
         }

         if (ArgusDstUserDataLen > 10) len++;
         sprintf (&buf[strlen(buf)], "%*sdstUdata%*s ", (len)/2, " ", (len)/2, " ");
         if (len & 0x01)
            sprintf(&buf[strlen(buf)], " ");
      }
   } else 
      sprintf (&buf[strlen(buf)], " srcUdata%cdstUdata ", RaFieldDelimiter);
}

void
ArgusPrintTCPExtensionsLabel (char *buf)
{
}

void
ArgusPrintSrcRateLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], RaSrcRateOptionLabel);
}

void
ArgusPrintDstRateLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], RaDstRateOptionLabel);
}

void
ArgusPrintRateLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], RaRateOptionLabel);
}

void
ArgusPrintLossLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], RaLossOptionLabel);
}

void
ArgusPrintSrcLoadLabel (char *buf)
{
   if (Aflag)
      sprintf (&buf[strlen(buf)], RaSrcAppLoadOptionLabel);
   else
      sprintf (&buf[strlen(buf)], RaSrcLoadOptionLabel);
}

void
ArgusPrintDstLoadLabel (char *buf)
{
   if (Aflag)
      sprintf (&buf[strlen(buf)], RaDstAppLoadOptionLabel);
   else
      sprintf (&buf[strlen(buf)], RaDstLoadOptionLabel);
}

void
ArgusPrintLoadLabel (char *buf)
{
   if (Aflag)
      sprintf (&buf[strlen(buf)], RaAppLoadOptionLabel);
   else
      sprintf (&buf[strlen(buf)], RaLoadOptionLabel);
}

void
ArgusPrintSrcTTLLabel (char *buf)
{
   if (RaPrintDistance)
      sprintf (&buf[strlen(buf)], "SrcDis ");
   else
      sprintf (&buf[strlen(buf)], "SrcTTL ");
}

void
ArgusPrintDstTTLLabel (char *buf)
{
   if (RaPrintDistance)
      sprintf (&buf[strlen(buf)], "DstDis ");
   else
      sprintf (&buf[strlen(buf)], "DstTTL ");
}

void
ArgusPrintMPLSLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], "  sMPLS    dMPLS  ");
}

void
ArgusPrintVLANLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], " sVLAN  dVLAN ");
}

void
ArgusPrintSrcTOSLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], "SrcTOS ");
}

void
ArgusPrintDstTOSLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], "DstTOS ");
}

void
ArgusPrintWindowLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], "SrcWin DstWin ");
}

void
ArgusPrintDurationLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], RaPrecisionPad);
   sprintf (&buf[strlen(buf)], "   Dur   ");
 
   sprintf (&buf[strlen(buf)], RaPrecisionPad);
   if (pflag) {
      sprintf (&buf[strlen(buf)], " ");
      if (pflag & 0x01)
         sprintf (&buf[strlen(buf)], " ");
   }
}

void
ArgusPrintAvgDurationLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], RaPrecisionPad);
   sprintf (&buf[strlen(buf)], "  AvgDur ");
 
   sprintf (&buf[strlen(buf)], RaPrecisionPad);
   if (pflag) {
      sprintf (&buf[strlen(buf)], " ");
      if (pflag & 0x01)
         sprintf (&buf[strlen(buf)], " ");
   }
}

void
ArgusPrintTransactionsLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], "Trans ");
}

void
ArgusPrintSequenceNumberLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], "  Seq  ");
}

void
ArgusPrintBinNumberLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], "  Bin  ");
}

void
ArgusPrintBinsLabel (char *buf)
{
   sprintf (&buf[strlen(buf)], " Bins  ");
}


void
RaSortQueue (struct RaQueueStruct *queue)
{
   int i = 0, cnt = queue->count;
   struct ArgusQueueHeader *qhdr;

   if (queue->array != NULL) {
      ArgusFree(queue->array);
      queue->array = NULL;
   }

   if (cnt > 0) {
      if ((queue->array = (struct ArgusQueueHeader **)
                  ArgusCalloc(sizeof(struct ArgusQueueHeader *), cnt + 1)) != NULL) {
         qhdr = queue->start;
         for (i = 0; i < cnt; i++) {
            queue->array[i] = qhdr;
            qhdr = qhdr->nxt;
            if (qhdr == queue->start)
               break;
         }

         qsort ((char *) queue->array, i + 1, sizeof (struct ArgusQueueHeader *), RaSortRoutine);

      } else
         ArgusLog (LOG_ERR, "RaSortQueue: ArgusCalloc(%d, %d) %s\n", sizeof(struct ArgusRecord *),
                                                                     cnt, strerror(errno));
   }
#ifdef ARGUSDEBUG
   ArgusDebug (3, "RaSortQueue(0x%x) returned\n", queue);
#endif
}

int
RaSortRoutine (const void *void1, const void *void2)
{
   int retn = 0, i = 0;
   struct ArgusRecordStore *store1 = *(struct ArgusRecordStore **)void1;
   struct ArgusRecordStore *store2 = *(struct ArgusRecordStore **)void2;
   struct ArgusRecord *argus1 = NULL;
   struct ArgusRecord *argus2 = NULL;

   if ((store1 && store1->data[0] && store1->data[0]->argus) && 
      ((store2 && store2->data[0] && store2->data[0]->argus))) {

      argus1 = store1->data[0]->argus;
      argus2 = store2->data[0]->argus;

      for (i = 0; i < ARGUS_MAX_SORT_ALG; i++)
         if (RaSortAlgorithms[i] != NULL)
            if ((retn = RaSortAlgorithms[i](argus2, argus1)) != 0)
               break;
   }

   return (retn);
}


int RaSortSrcId (struct ArgusRecord *a2, struct ArgusRecord *a1)
{
   unsigned int sid1 = a1->ahdr.argusid;
   unsigned int sid2 = a2->ahdr.argusid;

   return ((sid1 > sid2) ? 1 : ((sid1 == sid2) ? 0 : -1));
}


int RaSortStartTime (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   unsigned int retn;

   if ((retn = a2->argus_far.time.start.tv_sec - a1->argus_far.time.start.tv_sec) == 0)
      retn = a2->argus_far.time.start.tv_usec - a1->argus_far.time.start.tv_usec;

   return (retn);
}

int RaSortLastTime (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   unsigned int retn;

   if ((retn = a2->argus_far.time.last.tv_sec - a1->argus_far.time.last.tv_sec) == 0)
      retn = a2->argus_far.time.last.tv_usec - a1->argus_far.time.last.tv_usec;

   return (retn);
}

int RaSortCount (struct ArgusRecord *a2, struct ArgusRecord *a1)
{
   int cnt1 = 1, cnt2 = 1;
   struct ArgusAGRStruct *agr1, *agr2;
   struct ArgusFarHeaderStruct *a1farhdrs[32];
   struct ArgusFarHeaderStruct *a2farhdrs[32];

   ArgusIndexRecord (a1,  a1farhdrs);
   ArgusIndexRecord (a2,  a2farhdrs);

   agr1 = (struct ArgusAGRStruct *) a1farhdrs[ARGUS_AGR_DSR_INDEX];
   agr2 = (struct ArgusAGRStruct *) a2farhdrs[ARGUS_AGR_DSR_INDEX];

   if (agr1 != NULL)
      cnt1 = agr1->count;

   if (agr2 != NULL)
      cnt2 = agr2->count;
   
   return (cnt2 - cnt1);
}

int RaSortAvgDuration (struct ArgusRecord *a2, struct ArgusRecord *a1)
{
   return ((int)(RaGetuSecAvgDuration(a1) - RaGetuSecAvgDuration(a2)));
}

int RaSortDuration (struct ArgusRecord *a2, struct ArgusRecord *a1)
{
   return ((int)(RaGetuSecDuration(a1) - RaGetuSecDuration(a2)));
}

int RaSortSrcAddr (struct ArgusRecord *a2, struct ArgusRecord *a1)
{
  int retn = 0;

   if ((retn = (a1->ahdr.status & 0xFFFF) - (a2->ahdr.status & 0xFFFF)) == 0) {
      switch (a1->ahdr.status & 0xFFFF) {
         case ETHERTYPE_IP: {
            unsigned int addr1 = a1->argus_far.flow.ip_flow.ip_src;
            unsigned int addr2 = a2->argus_far.flow.ip_flow.ip_src;
            retn = (addr1 > addr2) ? 1 : ((addr1 == addr2) ? 0 : -1);
            break;
         }

         case ETHERTYPE_ARP: {
            unsigned int addr1 = a1->argus_far.flow.arp_flow.arp_spa;
            unsigned int addr2 = a2->argus_far.flow.arp_flow.arp_spa;
            retn = (addr1 > addr2) ? 1 : ((addr1 == addr2) ? 0 : -1);
            break;
         }

         default:
            retn = bcmp ((char *)&a2->argus_far.flow.mac_flow.ehdr.ether_shost, (char *)&a1->argus_far.flow.mac_flow.ehdr.ether_shost, 6);
            break;
      }
   }

   return (retn);
}

int RaSortDstAddr (struct ArgusRecord *a2, struct ArgusRecord *a1)
{
   int retn = 0;

   if ((retn = (a1->ahdr.status & 0xFFFF) - (a2->ahdr.status & 0xFFFF)) == 0) {
      switch (a1->ahdr.status & 0xFFFF) {
         case ETHERTYPE_IP: {
            unsigned int addr1 = a1->argus_far.flow.ip_flow.ip_dst;
            unsigned int addr2 = a2->argus_far.flow.ip_flow.ip_dst;
            retn = (addr1 > addr2) ? 1 : ((addr1 == addr2) ? 0 : -1);
            break;
         }

         case ETHERTYPE_ARP: {
            unsigned int addr1 = a1->argus_far.flow.arp_flow.arp_tpa;
            unsigned int addr2 = a2->argus_far.flow.arp_flow.arp_tpa;
            retn = (addr1 > addr2) ? 1 : ((addr1 == addr2) ? 0 : -1);
            break;
         }

         default:
            retn = bcmp ((char *)&a2->argus_far.flow.mac_flow.ehdr.ether_dhost, (char *)&a1->argus_far.flow.mac_flow.ehdr.ether_dhost, 6);
            break;
      }
   }

   return (retn);
}

int RaSortProtocol (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   int retn;

   if ((retn = (a1->ahdr.status & 0xFFFF) - (a2->ahdr.status & 0xFFFF)) == 0)
      if (((a1->ahdr.status & 0xFFFF) == ETHERTYPE_IP) && ((a2->ahdr.status & 0xFFFF) == ETHERTYPE_IP))
         if ((retn = a1->argus_far.flow.ip_flow.ip_p - a2->argus_far.flow.ip_flow.ip_p) == 0)
            retn = a1->argus_far.flow.ip_flow.tp_p - a2->argus_far.flow.ip_flow.tp_p;

   return (retn);
}

int RaSortSrcPort (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   return (a2->argus_far.flow.ip_flow.sport - a1->argus_far.flow.ip_flow.sport);
}

int RaSortDstPort (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   return (a2->argus_far.flow.ip_flow.dport - a1->argus_far.flow.ip_flow.dport);
}

int RaSortSrcTos (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   int retn = 0;

   if ((a1->ahdr.status & 0xFFFF) == ETHERTYPE_IP)
      retn = (a2->argus_far.attr_ip.stos - a1->argus_far.attr_ip.stos);

   return (retn);
}

int RaSortDstTos (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   int retn = 0;

   if ((a1->ahdr.status & 0xFFFF) == ETHERTYPE_IP)
      retn =  (a2->argus_far.attr_ip.dtos - a1->argus_far.attr_ip.dtos);

   return (retn);
}

int RaSortSrcTtl (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   int retn = 0;

   if ((a1->ahdr.status & 0xFFFF) == ETHERTYPE_IP)
      retn =  (a2->argus_far.attr_ip.sttl - a1->argus_far.attr_ip.sttl);

   return (retn);
}

int RaSortDstTtl (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   int retn = 0;

   if ((a1->ahdr.status & 0xFFFF) == ETHERTYPE_IP)
      retn = (a2->argus_far.attr_ip.dttl - a1->argus_far.attr_ip.dttl);

   return (retn);
}

int RaSortByteCount (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   long long cnt1, cnt2;
   int retn;

   if (Aflag) {
      cnt1 = a1->argus_far.src.appbytes + a1->argus_far.dst.appbytes;
      cnt2 = a2->argus_far.src.appbytes + a2->argus_far.dst.appbytes;
      retn = (int) (cnt1 - cnt2);

   } else {
      cnt1 = a1->argus_far.src.bytes + a1->argus_far.dst.bytes;
      cnt2 = a2->argus_far.src.bytes + a2->argus_far.dst.bytes;
      retn = (int) (cnt1 - cnt2);
   }

   return (retn);
}

int RaSortLoad (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   int src1, src2, dst1, dst2, retn = 0;
   struct timeval *start = NULL, *last = NULL;
   struct timeval  timebuf,  *time = &timebuf;
   double sec1, sec2;

   start = &a1->argus_far.time.start;
   last  = &a1->argus_far.time.last;
 
   *time = *last;
   time->tv_sec  -= start->tv_sec;
   time->tv_usec -= start->tv_usec;
 
   if (time->tv_usec < 0) {
      time->tv_sec--;
      time->tv_usec += 1000000;
   }
 
   if (time->tv_usec >= 1000000) {
      time->tv_sec++;
      time->tv_usec -= 1000000;
   }

   sec1 = (double)(time->tv_sec * 1.0) + (double)(time->tv_usec/1000000.0);

   start = &a2->argus_far.time.start;
   last  = &a2->argus_far.time.last;
 
   *time = *last;
   time->tv_sec  -= start->tv_sec;
   time->tv_usec -= start->tv_usec;
 
   if (time->tv_usec < 0) {
      time->tv_sec--;
      time->tv_usec += 1000000;
   }
 
   if (time->tv_usec >= 1000000) {
      time->tv_sec++;
      time->tv_usec -= 1000000;
   }

   sec2 = (double)(time->tv_sec * 1.0) + (double)(time->tv_usec/1000000.0);

   if (Aflag) {
      src1 = a1->argus_far.src.appbytes;
      dst1 = a1->argus_far.dst.appbytes;
      src2 = a2->argus_far.src.appbytes;
      dst2 = a2->argus_far.dst.appbytes;

   } else {
      src1 = a1->argus_far.src.bytes;
      dst1 = a1->argus_far.dst.bytes;
      src2 = a2->argus_far.src.bytes;
      dst2 = a2->argus_far.dst.bytes;
   }

   if (sec1 && sec2) {
      retn = (((double)(((src1 + dst1) * 1.0)/(double)(sec1 * 1.0)) >
               (double)(((src2 + dst2) * 1.0)/(double)(sec2 * 1.0))) ? 1 :
              ((double)(((src1 + dst1) * 1.0)/(double)(sec1 * 1.0)) ==
               (double)(((src2 + dst2) * 1.0)/(double)(sec2 * 1.0))) ? 0 : -1);
   } else {
      if (sec1)
         retn = 1;
      if (sec2)
         retn = -1;
   }

   return (retn);
}


int RaSortLoss (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
/*
struct ArgusTCPObjectMetrics {
   unsigned int seqbase, ackbytes;
   unsigned int bytes, rpkts;
   unsigned short win;
   unsigned char flags, pad;
};

struct ArgusTCPObject {
   unsigned char type, length;
   unsigned short status;
   unsigned int state;
   unsigned int options;
   unsigned int synAckuSecs, ackDatauSecs;
   struct ArgusTCPObjectMetrics src, dst;
};
*/
   int retn = 0;
   double a1pkts, a1rpkts, a1loss;
   double a2pkts, a2rpkts, a2loss;

   if ((a1->argus_far.flow.ip_flow.ip_p == IPPROTO_TCP) &&
       (a2->argus_far.flow.ip_flow.ip_p == IPPROTO_TCP)) {
      unsigned int a1DSRStatus = 0, a2DSRStatus = 0;
      struct ArgusFarHeaderStruct *a1farhdr[32], *a2farhdr[32];
      struct ArgusTCPObject *tcp1 = NULL;
      struct ArgusTCPObject *tcp2 = NULL;

      a1DSRStatus = ArgusIndexRecord (a1, a1farhdr);
      a2DSRStatus = ArgusIndexRecord (a2, a2farhdr);

      if ((a1DSRStatus & ARGUS_TCP_DSR_STATUS) && (a2DSRStatus & ARGUS_TCP_DSR_STATUS)) {
         tcp1 = (struct ArgusTCPObject *) a1farhdr[ARGUS_TCP_DSR_INDEX];
         tcp2 = (struct ArgusTCPObject *) a2farhdr[ARGUS_TCP_DSR_INDEX];

         a1pkts  = (a1->argus_far.src.count + a1->argus_far.dst.count) * 1.0;
         a2pkts  = (a2->argus_far.src.count + a2->argus_far.dst.count) * 1.0;
         a1rpkts = (tcp1->src.rpkts + tcp1->dst.rpkts) * 1.0;
         a2rpkts = (tcp2->src.rpkts + tcp2->dst.rpkts) * 1.0;

         a1loss = a1rpkts/a1pkts;
         a2loss = a2rpkts/a2pkts;

         retn = (a1loss > a2loss) ? 1 : ((a1loss == a2loss) ? 0 : -1);
      }
   }

   return (retn);
}


int
RaSortRate (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   int src1 = 0, src2 = 0, dst1 = 0, dst2 = 0, retn = 0;
   struct timeval *start = NULL, *last = NULL;
   struct timeval  timebuf,  *time = &timebuf;
   double sec1, sec2;

   start = &a1->argus_far.time.start;
   last  = &a1->argus_far.time.last;
 
   *time = *last;
   time->tv_sec  -= start->tv_sec;
   time->tv_usec -= start->tv_usec;
 
   if (time->tv_usec < 0) {
      time->tv_sec--;
      time->tv_usec += 1000000;
   }
 
   if (time->tv_usec >= 1000000) {
      time->tv_sec++;
      time->tv_usec -= 1000000;
   }

   sec1 = (double)(time->tv_sec * 1.0) + (double)(time->tv_usec/1000000.0);

   start = &a2->argus_far.time.start;
   last  = &a2->argus_far.time.last;
 
   *time = *last;
   time->tv_sec  -= start->tv_sec;
   time->tv_usec -= start->tv_usec;
 
   if (time->tv_usec < 0) {
      time->tv_sec--;
      time->tv_usec += 1000000;
   }
 
   if (time->tv_usec >= 1000000) {
      time->tv_sec++;
      time->tv_usec -= 1000000;
   }

   sec2 = (double)(time->tv_sec * 1.0) + (double)(time->tv_usec/1000000.0);

   if (a1->argus_far.src.count > 1)
      src1 = a1->argus_far.src.count;

   if (a1->argus_far.dst.count > 1)
      dst1 = a1->argus_far.dst.count;

   if (a2->argus_far.src.count > 1)
      src2 = a2->argus_far.src.count;

   if (a2->argus_far.dst.count > 1)
      dst2 = a2->argus_far.dst.count;

   if (sec1 && sec2) {
      retn = (((double)(((src1 + dst1) * 1.0)/(double)(sec1 * 1.0)) >
               (double)(((src2 + dst2) * 1.0)/(double)(sec2 * 1.0))) ? 1 :
              ((double)(((src1 + dst1) * 1.0)/(double)(sec1 * 1.0)) ==
               (double)(((src2 + dst2) * 1.0)/(double)(sec2 * 1.0))) ? 0 : -1);
   } else {
      if (sec1)
         retn = 1;
      if (sec2)
         retn = -1;
   }

   return (retn);
}

int
RaSortTranRef (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   return (a1->argus_far.ArgusTransRefNum - a2->argus_far.ArgusTransRefNum);
}

int
RaSortSeq (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   return (a1->ahdr.seqNumber - a2->ahdr.seqNumber);
}


int
RaSortSrcByteCount (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   if (Aflag)
      return (a1->argus_far.src.appbytes - a2->argus_far.src.appbytes);
   else
      return (a1->argus_far.src.bytes - a2->argus_far.src.bytes);
}

int
RaSortDstByteCount (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   if (Aflag)
      return (a1->argus_far.dst.appbytes - a2->argus_far.dst.appbytes);
   else
      return (a1->argus_far.dst.bytes - a2->argus_far.dst.bytes);
}


int
RaSortPktsCount (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   return ((a1->argus_far.src.count + a1->argus_far.dst.count) -
           (a2->argus_far.src.count + a2->argus_far.dst.count));
}

int
RaSortSrcPktsCount (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   return (a1->argus_far.src.count - a2->argus_far.src.count);
}

int
RaSortDstPktsCount (struct ArgusRecord *a1, struct ArgusRecord *a2)
{
   return (a1->argus_far.dst.count - a2->argus_far.dst.count);
}


char RaUserDataStr[MAXSTRLEN];

char *
RaGetUserDataString (struct ArgusRecord *argus)
{
   char *retn = RaUserDataStr;
   char strbuf[MAXSTRLEN], *str = strbuf;
   char delim = ' ';
   int len = 0;

   bzero (RaUserDataStr, MAXSTRLEN);

   if ((RaFieldDelimiter != ' ') && (RaFieldDelimiter != '\0'))
      delim = RaFieldDelimiter;

   if (ArgusSrcUserDataLen > 0) {
      len = 0;
      if (ArgusThisFarStatus & ARGUS_SRCUSRDATA_DSR_STATUS) {
         struct ArgusUserStruct *user = (struct ArgusUserStruct *) ArgusThisFarHdrs[ARGUS_SRCUSRDATA_DSR_INDEX];

         len = (user->length - 1) * 4;
         len = (len < argus->argus_far.src.appbytes) ? len : argus->argus_far.src.appbytes;
         len = len > ArgusSrcUserDataLen ? ArgusSrcUserDataLen : len;
   
         bzero(strbuf, sizeof(strbuf));
         if ((len = ArgusEncode (&user->data, NULL, len, str, sizeof(strbuf))) != 0)
            sprintf (RaUserDataStr, "%cs[%d]=%s", delim, len, str);
   
      } else
         if (delim != ' ')
            sprintf (&RaUserDataStr[strlen(RaUserDataStr)], "%c", delim);

      if (delim == ' ')
         sprintf (&RaUserDataStr[strlen(RaUserDataStr)], "%*s", (ArgusSrcUserDataLen - len) + 1, " ");
   }


   if (ArgusDstUserDataLen > 0) {
      len = 0;
      if (ArgusThisFarStatus & ARGUS_DSTUSRDATA_DSR_STATUS) {
         struct ArgusUserStruct *user = (struct ArgusUserStruct *) ArgusThisFarHdrs[ARGUS_DSTUSRDATA_DSR_INDEX];

         len = (user->length - 1) * 4;
         len = (len < argus->argus_far.dst.appbytes) ? len : argus->argus_far.dst.appbytes;
         len = len > ArgusDstUserDataLen ? ArgusDstUserDataLen : len;
   
         bzero(strbuf, sizeof(strbuf));
         if ((len = ArgusEncode (&user->data, NULL, len, str, sizeof(strbuf))) != 0)
            sprintf (&RaUserDataStr[strlen(RaUserDataStr)], "%cd[%d]=%s", delim, len, str);
      } else
         if (delim != ' ')
            sprintf (&RaUserDataStr[strlen(RaUserDataStr)], "%c", delim);
   }

   return (retn);
}


/*
   Format is "bins[L]:range"
   where range is  value-value
   value is %d[ums]
*/

int RaHistoTimeParse (char *);
int *RaHistoTimeValues = NULL;
int RaHistoStart = 0;
int RaHistoEnd = 0;
int RaHistoBins = 0;

int
RaHistoTimeParse (char *str) 
{
   int RaHistoTimeLog = 0;
   double RaHistoBinSize = 0.0, RaHistoLogInterval = 0.0;
   double RaHistoStartLog = 0.0, RaHistoEndLog = 0.0;
   int retn = 0, i;
   char *ptr, *vptr, tmpbuf[64], *tmp = tmpbuf;

   bzero (tmpbuf, 64);
   strcpy (tmpbuf, str);

   if ((ptr = strchr (tmp, ':')) != NULL) {
      *ptr++ = '\0';
      vptr = ptr;

      if (strchr (tmp, 'L'))
         RaHistoTimeLog++;

      if (isdigit((int)*tmp))
         if ((RaHistoBins = atoi(tmp)) < 0)
            return (retn);

      if ((ptr = strchr (vptr, '-')) != NULL) {
         *ptr++ = '\0';
         if ((RaHistoStart = atoi(vptr)) < 0)
            return (retn);
         if ((RaHistoEnd = atoi(ptr)) < 0)
            return (retn);

         while (isdigit((int)*ptr)) ptr++;
         switch (*ptr) {
            case 'u': RaHistoStart *= 1;
                      RaHistoEnd   *= 1;        break;
            case 'm': RaHistoStart *= 1000;
                      RaHistoEnd   *= 1000;     break;
            case 's': RaHistoStart *= 1000000;
                      RaHistoEnd   *= 1000000;  break;
            case 'M': RaHistoStart *= 60000000;
                      RaHistoEnd   *= 60000000; break;
            case '\0': break;

            default:
               return (retn);
         }

         retn++;

         RaHistoTimeSeries = RaHistoBins;

         if ((RaHistoTimeValues = (int *) ArgusCalloc (RaHistoBins + 1, sizeof(int))) != NULL) {
            for (i = 0; i < RaHistoBins; i++) {
               if (RaHistoTimeLog) {
                  if (i == 0) {
                     if (RaHistoStart > 0)
                        RaHistoStartLog = log10(RaHistoStart);

                     RaHistoEndLog = log10(RaHistoEnd);

                     RaHistoLogInterval = ((RaHistoEndLog - RaHistoStartLog)/(RaHistoBins * 1.0));
                  }

                  RaHistoTimeValues[i] = RaHistoStart + pow(10.0, RaHistoLogInterval * i);

               } else {
                  if (i == 0)
                     RaHistoBinSize = ((RaHistoEnd - RaHistoStart) * 1.0) / RaHistoBins * 1.0;

                  RaHistoTimeValues[i] = RaHistoStart + RaHistoBinSize * (i + 1);
               }
            }

            RaHistoTimeValues[RaHistoBins] = RaHistoEnd;

         } else
            ArgusLog (LOG_ERR, "RaHistoTimeParse: ArgusCalloc %s\n", strerror(errno));
      }
   }

   return (retn);
}
